// created: 8.6.01 modified: 24.8.01 final clean-up: 24.4.02
// Stephan Dale
// MAVERIK (v6.2) avatar library

#include <stdlib.h>
#include <math.h>
#include "maverik.h"
#include "mav_avatar.h" // for avatar datatypes
#include "mavlib_avatar.h" // for parts definitions
#include "mav_tdm.h"

#include "sjd_backprop.h" // for the NN datatype
#include "avatarlib.h"
#include "netlib.h"

#define RIGHT_ARM 0
#define LEFT_ARM 1
#define TRACKER_SCALE 0.01 // cm to m

/**************************************************************************************************/

/********************************/
/* internal function prototypes */
/********************************/
void getPY(MAV_vector vec, float *pitch, float *yaw);
MAV_vector flipHand(MAV_vector handPos);
void flipPY(float pitch, float yaw, float *newPitch, float *newYaw);
MAV_matrix setScalingMatrix(float x, float y, float z);

void positionAvatar(void *unused);
void setArm(int arm);

/**************************************************************************************************/

/**********************************************/
/* globals (that alter the avatar's position) */
/**********************************************/
MAV_avatar *avatar; // pointer to the avatar
BPNN *net; // pointer to the NN

int holdingLeft = 1;
int holdingRight = 1;
MAV_matrix backMatrix; // gives position and rotation of back in world co-ordinates.
MAV_matrix rightHandMatrix, leftHandMatrix; // position and rotation of hands in world co-ordinates.

// avatar positional data
MAV_vector leftHand, rightHand; // the position of the avatars hands
float avatarYaw, headRoll, headPitch, headYaw, torsoRoll, torsoPitch, torsoYaw;
//float leftHandRollWorld, leftHandPitchWorld, leftHandYawWorld, rightHandRollWorld, rightHandPitchWorld, rightHandYawWorld;
float leftHandRoll, leftHandPitch, leftHandYaw, rightHandRoll, rightHandPitch, rightHandYaw;
MAV_vector *eyePoint;
float *leftArmRoll, *leftArmPitch, *leftArmYaw, *rightArmRoll, *rightArmPitch, *rightArmYaw;

// avatar data
MAV_vector eyePointAvatar; // eye point relative to upper torso
// keyboard data
float leftArmRollKeyboard, leftArmPitchKeyboard, leftArmYawKeyboard, rightArmRollKeyboard, rightArmPitchKeyboard, rightArmYawKeyboard;
// tracked data
MAV_vector eyePointTracked; // eye point relative to back sensor
float leftArmRollTracked, leftArmPitchTracked, leftArmYawTracked, rightArmRollTracked, rightArmPitchTracked, rightArmYawTracked;
// neural network data
float leftArmRollNetwork, leftArmPitchNetwork, leftArmYawNetwork, rightArmRollNetwork, rightArmPitchNetwork, rightArmYawNetwork;

/*****************/
/* other globals */
/*****************/

MAV_vector avatarDirection; // the direction that the avatar (legs) is facing

// offsets and scaling
MAV_vector eyeOffset, handOffset, upperArmOffset, torsoOffset;
MAV_matrix torsoCorrection; // matrix applied to correct initial torso position
float shoulderWidth, upperArmLength, lowerArmLength; // dimensions in meters
MAV_matrix torsoScale, upperArmScale, lowerArmScale; // the scaling matrices in force

// tracker -> avatar co-ordinate translations
MAV_matrix headSATrans, leftHandSATrans, rightHandSATrans, torsoSATrans;
// variables needed for calculating arm position (only global to allow the show* functions to display them)
MAV_vector predictedUpperArm, upperArm, lowerArm, hand, xAxis, yAxis, zAxis;

/**************************************************************************************************/

/*******************************************************************************************/
/* return pitch and yaw of upper arm (-Y vector) relative to shoulder (assuming roll is 0) */
/*******************************************************************************************/
void getPY(MAV_vector vec, float *pitch, float *yaw)
 {
 MAV_vector v, normal;
 float cosAngle;

 *pitch = -MAV_RAD2DEG(acos( mav_vectorDotProduct(mav_vectorScalar(MAV_Y_VECTOR,-1),mav_vectorNormalise(vec)) ));
 *pitch += 360;

 v = mav_vectorSet(vec.x, 0, vec.z); // reflect onto the xz plane
 *yaw = MAV_RAD2DEG(acos( mav_vectorDotProduct(MAV_Z_VECTOR,mav_vectorNormalise(v)) )); // get yaw of arm in xz plane relative to z-vector
 normal = mav_vectorCrossProduct(MAV_Z_VECTOR,mav_vectorNormalise(v)); // get normal to z-vector
 cosAngle = mav_vectorDotProduct(mav_vectorNormalise(normal),MAV_Y_VECTOR);
 if ( cosAngle < -0.999999)  // if normal at 180 degrees to y-vector, then negative yaw
   *yaw = -*yaw + 360;

 }

/***********************************************/
/* functions to flip arm data for opposite arm */
/***********************************************/
MAV_vector flipHand(MAV_vector handPos)
 {
 handPos.x = -handPos.x; // reflect in YZ axis
 return handPos;
 }

void flipPY(float pitch, float yaw, float *newPitch, float *newYaw)
 {
 MAV_vector vec;

 vec = mav_vectorMult(MAV_Y_VECTOR, mav_matrixRPYSet(MAV_ID_MATRIX, 0, pitch, yaw));
 vec = mav_vectorScalar(vec, -1);
 vec = flipHand(vec);
 getPY(vec, newPitch, newYaw);
 }

/**************************************************************************************************/

/*************************/
/* form a scaling matrix */
/*************************/
MAV_matrix setScalingMatrix(float x, float y, float z)
 {
 MAV_matrix mat;

 // form a scaling matrix

 mat.mat[0][0] = x;
 mat.mat[0][1] = 0.0;
 mat.mat[0][2] = 0.0;
 mat.mat[0][3] = 0.0;

 mat.mat[1][0] = 0.0;
 mat.mat[1][1] = y;
 mat.mat[1][2] = 0.0;
 mat.mat[1][3] = 0.0;

 mat.mat[2][0] = 0.0;
 mat.mat[2][1] = 0.0;
 mat.mat[2][2] = z;
 mat.mat[2][3] = 0.0;

 mat.mat[3][0] = 0.0;
 mat.mat[3][1] = 0.0;
 mat.mat[3][2] = 0.0;
 mat.mat[3][3] = 1.0;

 return mat;
 }

/**************************************************************************************************/

/**************************************/
/* show vectors calculated in setArms */
/**************************************/
void showSetArmVectors(void *unused)
 {
 MAV_matrix mat;
 float r,p,y;
 MAV_vector vec, shoulder, upperArmPos;

 // get avatar shoulder point
 mat = backMatrix;
 mat = mav_matrixMult(mat, avatar->part[RIGHT_CLAVICLE]->position);
 mat = mav_matrixMult(mat, avatar->part[RIGHT_CLAVICLE]->rotation);
 mat = mav_matrixMult(mat, avatar->part[RIGHT_UPPER_ARM]->position);
 shoulder = mav_matrixXYZGet(mat);

 // line from shoulder to hand (hand vector)
 mav_gfxLineBegin();
  mav_gfxVertex(shoulder);
  mav_gfxVertex(mav_vectorMult(hand, mat));
 mav_gfxLineEnd();

 // line from shoulder to predicted elbow point (predictedUpperArm vector)
 mav_gfxLineBegin();
  mav_gfxVertex(shoulder);
  mav_gfxVertex(mav_vectorMult(predictedUpperArm, mat));
 mav_gfxLineEnd();

 upperArmPos = mav_vectorMult(upperArm, mat);

 // line from shoulder to corrected elbow point (upperArm vector)
 mav_gfxLineBegin();
  mav_gfxVertex(shoulder);
  mav_gfxVertex(upperArmPos);
 mav_gfxLineEnd();

 mat = mav_matrixMult(mat, avatar->part[RIGHT_UPPER_ARM]->rotation);
 mat = mav_matrixXYZSet(mat, mav_vectorSet(0,0,0));

 // line from (corrected) elbow point to hand (lowerArm vector)
 mav_gfxLineBegin();
  mav_gfxVertex(upperArmPos);
  mav_gfxVertex(mav_vectorAdd(upperArmPos, mav_vectorMult(lowerArm, mat)));
 mav_gfxLineEnd();
 }

/***************************/
/* show upper body vectors */
/***************************/
void showVectors(void *unused)
 {
 MAV_matrix mat;
 float r,p,y;
 MAV_vector vec, back, shoulder, upperArmPos;

 // get avatar back point
 mat = backMatrix;
 back = mav_matrixXYZGet(mat);

 // line from back to hand
 mav_gfxLineBegin();
  mav_gfxVertex(back);
  mav_gfxVertex(mav_vectorMult(mav_vectorMult(rightHand, mav_matrixInverse(torsoScale)), mat));
 mav_gfxLineEnd();

 // get shoulder point
 vec = mav_vectorMult(mav_vectorSet(-shoulderWidth,0,0), mav_matrixInverse(torsoScale)); // undo torso scaling
 mat = mav_matrixMult(mat, mav_matrixXYZSet(MAV_ID_MATRIX, vec));
 shoulder = mav_matrixXYZGet(mat);

 mav_gfxLineBegin();
  mav_gfxVertex(back);
  mav_gfxVertex(shoulder);
 mav_gfxLineEnd();

 // set upperArm position relative to back rotation
 vec = mav_vectorMult(MAV_Y_VECTOR, mav_matrixRPYSet(MAV_ID_MATRIX, *rightArmRoll, *rightArmPitch, *rightArmYaw));
 vec = mav_vectorScalar(vec, -upperArmLength/mav_vectorMag(vec));

 // line from shoulder to hand (hand vector)
 mav_gfxLineBegin();
  mav_gfxVertex(shoulder);
  mav_gfxVertex(mav_vectorMult(mav_vectorMult(vec, mav_matrixInverse(torsoScale)), mat));
 mav_gfxLineEnd();

 }

/************************************/
/* show NN predicted elbow position */
/************************************/
void showNetworkElbow(void *unused)
 {
 MAV_matrix mat, elbow;
 MAV_vector vec, shoulder;

 // get avatar back matrix
 mat = backMatrix;

 // move to shoulder point
 vec = mav_vectorMult(mav_vectorSet(-shoulderWidth,0,0),mav_matrixInverse(torsoScale)); // undo torso scaling
 mat = mav_matrixMult(mat, mav_matrixXYZSet(MAV_ID_MATRIX, vec));
 shoulder = mav_matrixXYZGet(mat);

 // set upperArm position relative to back rotation
 vec = mav_vectorMult(MAV_Y_VECTOR, mav_matrixRPYSet(MAV_ID_MATRIX, rightArmRollNetwork, rightArmPitchNetwork, rightArmYawNetwork));
 vec = mav_vectorScalar(vec, -upperArmLength/mav_vectorMag(vec)); // ensure correct length

 // draw predicted upper arm
 mav_gfxLineBegin();
  mav_gfxVertex(shoulder);
  mav_gfxVertex(mav_vectorMult(mav_vectorMult(vec, mav_matrixInverse(torsoScale)), mat));
 mav_gfxLineEnd();

 }

/*******************************/
/* show tracked elbow position */
/*******************************/
void showTrackedElbow(void *unused)
 {
 MAV_matrix mat, elbow;
 MAV_vector vec, shoulder;

 // get avatar back matrix
 mat = backMatrix;

 // move to shoulder point
 vec =  mav_vectorMult(mav_vectorSet(-shoulderWidth,0,0), mav_matrixInverse(torsoScale)); // undo torso scaling
 mat = mav_matrixMult(mat, mav_matrixXYZSet(MAV_ID_MATRIX, vec));
 shoulder = mav_matrixXYZGet(mat);

 // set upper arm relative to shoulder
 vec = mav_vectorMult(MAV_Y_VECTOR, mav_matrixRPYSet(MAV_ID_MATRIX, rightArmRollTracked, rightArmPitchTracked, rightArmYawTracked));
 vec = mav_vectorScalar(vec, -upperArmLength/mav_vectorMag(vec)); // ensure correct length

 // draw tracked upper arm
 mav_gfxLineBegin();
  mav_gfxVertex(shoulder);
  mav_gfxVertex(mav_vectorMult(mav_vectorMult(vec, mav_matrixInverse(torsoScale)), mat));
 mav_gfxLineEnd();

 }

/******************************/
/* show axis of various parts */
/******************************/
void showAxis(void *unused)
 {
 MAV_matrix mat, mat1;
 MAV_vector back, shoulder, neck, arm, larm, hnd;
 MAV_vector x, y, z;

 // get back point
 mat = backMatrix;
 back = mav_matrixXYZGet(mat);

 // draw torso axis
 mav_gfxLineBegin();
  mav_gfxVertex(back);
  mav_gfxVertex(mav_vectorMult(mav_matrixXAxisGet(avatar->part[UPPER_TORSO]->rotation),mat));
 mav_gfxLineEnd();
 mav_gfxLineBegin();
  mav_gfxVertex(back);
  mav_gfxVertex(mav_vectorMult(mav_matrixYAxisGet(avatar->part[UPPER_TORSO]->rotation),mat));
 mav_gfxLineEnd();
 mav_gfxLineBegin();
  mav_gfxVertex(back);
  mav_gfxVertex(mav_vectorMult(mav_matrixZAxisGet(avatar->part[UPPER_TORSO]->rotation),mat));
 mav_gfxLineEnd();

 // get neck point
 mat1 = mav_matrixMult(mat, avatar->part[NECK]->position);
 mat1 = mav_matrixMult(mat1, avatar->part[NECK]->rotation);
 mat1 = mav_matrixMult(mat, avatar->part[HEAD]->position);
 neck = mav_matrixXYZGet(mat1);

 // get shoulder point
 mat = mav_matrixMult(mat, avatar->part[RIGHT_CLAVICLE]->position);
 mat = mav_matrixMult(mat, avatar->part[RIGHT_CLAVICLE]->rotation);
 mat = mav_matrixMult(mat, avatar->part[RIGHT_UPPER_ARM]->position);
 shoulder = mav_matrixXYZGet(mat);

 // draw upper arm axis
 mav_gfxLineBegin();
  mav_gfxVertex(shoulder);
  mav_gfxVertex(mav_vectorMult(mav_matrixXAxisGet(avatar->part[RIGHT_UPPER_ARM]->rotation),mat));
 mav_gfxLineEnd();
 mav_gfxLineBegin();
  mav_gfxVertex(shoulder);
  mav_gfxVertex(mav_vectorMult(mav_matrixYAxisGet(avatar->part[RIGHT_UPPER_ARM]->rotation),mat));
 mav_gfxLineEnd();
 mav_gfxLineBegin();
  mav_gfxVertex(shoulder);
  mav_gfxVertex(mav_vectorMult(mav_matrixZAxisGet(avatar->part[RIGHT_UPPER_ARM]->rotation),mat));
 mav_gfxLineEnd();

 // get lower arm point
 mat = mav_matrixMult(mat, avatar->part[RIGHT_UPPER_ARM]->rotation);
 mat = mav_matrixMult(mat, avatar->part[RIGHT_LOWER_ARM]->position);
 larm = mav_matrixXYZGet(mat);

 // draw lower arm axis
 mav_gfxLineBegin();
  mav_gfxVertex(larm);
  mav_gfxVertex(mav_vectorMult(mav_matrixXAxisGet(avatar->part[RIGHT_LOWER_ARM]->rotation),mat));
 mav_gfxLineEnd();
 mav_gfxLineBegin();
  mav_gfxVertex(larm);
  mav_gfxVertex(mav_vectorMult(mav_matrixYAxisGet(avatar->part[RIGHT_LOWER_ARM]->rotation),mat));
 mav_gfxLineEnd();
 mav_gfxLineBegin();
  mav_gfxVertex(larm);
  mav_gfxVertex(mav_vectorMult(mav_matrixZAxisGet(avatar->part[RIGHT_LOWER_ARM]->rotation),mat));
 mav_gfxLineEnd();

 // get hand point
 mat = mav_matrixMult(mat, avatar->part[RIGHT_LOWER_ARM]->rotation);
 mat = mav_matrixMult(mat, avatar->part[RIGHT_HAND]->position);
 hnd = mav_matrixXYZGet(mat);

 // draw hand axis
 mav_gfxLineBegin();
  mav_gfxVertex(hnd);
  mav_gfxVertex(mav_vectorMult(mav_matrixXAxisGet(avatar->part[RIGHT_HAND]->rotation),mat));
 mav_gfxLineEnd();
 mav_gfxLineBegin();
  mav_gfxVertex(hnd);
  mav_gfxVertex(mav_vectorMult(mav_matrixYAxisGet(avatar->part[RIGHT_HAND]->rotation),mat));
 mav_gfxLineEnd();
 mav_gfxLineBegin();
  mav_gfxVertex(hnd);
  mav_gfxVertex(mav_vectorMult(mav_matrixZAxisGet(avatar->part[RIGHT_HAND]->rotation),mat));
 mav_gfxLineEnd();
}

/**************************************************************************************************/

/*****************/
/* create avatar */
/*****************/
void createAvatar(MAV_SMS **sms)
 {
 MAV_BB bb;
 MAV_matrix mat;
 MAV_vector vec, shoulderPt, elbowPt, wristPt;
 printf("creating avatar...");

 // create avatar
 avatar = mav_avatarNew(NULL);

 // set eye point pointer
 eyePoint = &eyePointAvatar; // set in rendering-loop functions positionAvatar or trackUpperBody (used in avatarView view modifier)

 // set arm position pointers
 leftArmRoll = &leftArmRollKeyboard;
 leftArmPitch = &leftArmPitchKeyboard;
 leftArmYaw = &leftArmYawKeyboard;
 rightArmRoll = &rightArmRollKeyboard;
 rightArmPitch = &rightArmPitchKeyboard;
 rightArmYaw = &rightArmYawKeyboard;

 // turn my arm positioning on
 holdingLeft = 1;
 holdingRight = 1;
 avatar->holding_left = 0;
 avatar->holding_right = 0;

 // initialise world co-ordinate matrices
 backMatrix = MAV_ID_MATRIX; // these are set in the rendering-loop functino positionAvatar
 leftHandMatrix = MAV_ID_MATRIX;
 rightHandMatrix = MAV_ID_MATRIX;

 // initial scales
 torsoScale = MAV_ID_MATRIX;
 upperArmScale = MAV_ID_MATRIX;
 lowerArmScale = MAV_ID_MATRIX;

 // calculate shoulder point
 mat = mav_matrixMult(MAV_ID_MATRIX, avatar->part[RIGHT_CLAVICLE]->position);
 mat = mav_matrixMult(mat, avatar->part[RIGHT_CLAVICLE]->rotation);
 mat = mav_matrixMult(mat, avatar->part[RIGHT_UPPER_ARM]->position);
 shoulderPt = mav_matrixXYZGet(mat); // position of shoulder relative to back

 // default shoulder width
 shoulderWidth = fabs(shoulderPt.x);

 // default upper arm length
 mat = mav_matrixMult(mat, avatar->part[RIGHT_UPPER_ARM]->rotation);
 mat = mav_matrixMult(mat, avatar->part[RIGHT_LOWER_ARM]->position);
 elbowPt = mav_matrixXYZGet(mat);
 vec = mav_vectorSub(elbowPt, shoulderPt);
 upperArmLength = mav_vectorMag(vec);
 
 // default lower arm length 
 mat = mav_matrixMult(mat, avatar->part[RIGHT_LOWER_ARM]->rotation);
 mat = mav_matrixMult(mat, avatar->part[RIGHT_HAND]->position);
 wristPt = mav_matrixXYZGet(mat);
 vec = mav_vectorSub(wristPt, elbowPt);
 lowerArmLength = mav_vectorMag(vec);

 // set avatar posture
 resetPosition();

 // position avatar
 mav_frameFn1Add(positionAvatar, NULL);

 // add avatar to SMS
 mav_SMSObjectAdd(*sms, mav_objectNew(mav_class_avatar, avatar)); // do not need to give the address of avatar, as it is already a pointer

 printf("done\n");
 }

/*************************/
/* reset avatar position */
/*************************/

void resetPosition(void)
 {
 avatar->matrix = mav_matrixSet(0,0,0, 0,1.7,0);
 avatar->vertical = mav_matrixSet(0,0,0,0,-1.7,0); // as default
 avatar->rotation = mav_matrixSet(0,0,180, 0,0,0); // as default (makes avatar point down the z axis towards the origin)
// avatar->rotation = mav_matrixSet(0,0,0, 0,0,0);
 avatarYaw = 0;

 headRoll = 0;
 headPitch = 0;
 headYaw = 0;
 torsoRoll = 0;
 torsoPitch = 0;
 torsoYaw = 0;
 leftHandRoll = 0;
 leftHandPitch = 0;//-90;
 leftHandYaw = 0;
 rightHandRoll = 0;//180;
 rightHandPitch = 0;//-90;
 rightHandYaw = 0;

 // initialise network data
 leftArmRollKeyboard = 0;
 leftArmPitchKeyboard = 0;
 leftArmYawKeyboard = 0;
 rightArmRollKeyboard = 0;
 rightArmPitchKeyboard = 0;
 rightArmYawKeyboard = 0;

 // initialise tracker data
 leftArmRollTracked = 0;
 leftArmPitchTracked = 0;
 leftArmYawTracked = 0;
 rightArmRollTracked = 0;
 rightArmPitchTracked = 0;
 rightArmYawTracked = 0;

 leftHand = mav_vectorSet(shoulderWidth, -(upperArmLength+lowerArmLength), 0);
 rightHand = mav_vectorSet(-shoulderWidth, -(upperArmLength+lowerArmLength), 0);
 }

/************************************/
/* scale avatar to users dimensions */
/************************************/
int scaleAvatar(char *fileName)
 {
 FILE *in;
 MAV_BB bb;
 float newShoulderWidth = shoulderWidth;
 float newUpperArmLength = upperArmLength;
 float newLowerArmLength = lowerArmLength;
 float scale;
 MAV_matrix mat, scaleMatrix;
 MAV_vector claviclePt, shoulderPt, elbowPt, handPt;

 // read user's data file to get body proportions
 if ((in = fopen(fileName,"r"))!=NULL)
   {
   printf("reading user's dimensions from \"%s\"\n", fileName);
   if (!feof(in)){ fscanf(in, "%f", &newShoulderWidth); }
   if (!feof(in)){ fscanf(in, "%f", &newUpperArmLength); }
   if (!feof(in)){ fscanf(in, "%f", &newLowerArmLength); }
   fclose(in); // close file
   }
 else
   {
   printf("failed to open \"%s\" for reading\n", fileName);
   return(0);
   }

 // scale  torso
 scale = newShoulderWidth/shoulderWidth;
 scaleMatrix = setScalingMatrix(scale, scale, scale);
 torsoScale = mav_matrixMult(torsoScale, scaleMatrix); // update glodal torso scale
 // apply torso scaling to hips so that both torso and legs scale
 avatar->part[HIPS]->position = mav_matrixMult(avatar->part[HIPS]->position, scaleMatrix); // scale hips (torso and legs follow)

 // prevent neck and arms scaling with torso
 avatar->part[RIGHT_UPPER_ARM]->position = mav_matrixMult(avatar->part[RIGHT_UPPER_ARM]->position, mav_matrixInverse(scaleMatrix));
 avatar->part[LEFT_UPPER_ARM]->position = mav_matrixMult(avatar->part[LEFT_UPPER_ARM]->position, mav_matrixInverse(scaleMatrix));

 // scale upper arm
 scale = newUpperArmLength/upperArmLength;
 scaleMatrix = setScalingMatrix(scale, scale, scale);
 upperArmScale = mav_matrixMult(upperArmScale, scaleMatrix);
 avatar->part[RIGHT_UPPER_ARM]->position = mav_matrixMult(avatar->part[RIGHT_UPPER_ARM]->position, scaleMatrix); // scale upper arm
 avatar->part[LEFT_UPPER_ARM]->position = mav_matrixMult(avatar->part[LEFT_UPPER_ARM]->position, scaleMatrix);
 // prevent lower arm scaling with upper arm
 avatar->part[RIGHT_LOWER_ARM]->position = mav_matrixMult(avatar->part[RIGHT_LOWER_ARM]->position, mav_matrixInverse(scaleMatrix));
 avatar->part[LEFT_LOWER_ARM]->position = mav_matrixMult(avatar->part[LEFT_LOWER_ARM]->position, mav_matrixInverse(scaleMatrix));

 // scale lower arm
 scale = newLowerArmLength/lowerArmLength;
 scaleMatrix = setScalingMatrix(scale, scale, scale);
 lowerArmScale = mav_matrixMult(lowerArmScale, scaleMatrix);
 avatar->part[RIGHT_LOWER_ARM]->position = mav_matrixMult(avatar->part[RIGHT_LOWER_ARM]->position, scaleMatrix); // scale lower arm
 avatar->part[LEFT_LOWER_ARM]->position = mav_matrixMult(avatar->part[LEFT_LOWER_ARM]->position, scaleMatrix);

 // set new lengths
 shoulderWidth = newShoulderWidth;
 upperArmLength = newUpperArmLength;
 lowerArmLength = newLowerArmLength;

 /*printf("neck length = %f\n", neckLength);*/
 printf("shoulder width = %f\n", shoulderWidth);
 printf("upper arm length = %f\n", upperArmLength);
 printf("lower arm length = %f\n", lowerArmLength);

 return(1);
 }

/**************************************************************************************************/

/***************/
/* donkey kong */
/***************/
void donkeyKong(void)
 {
 MAV_matrix mat = setScalingMatrix(1.8,1.8,1.8);
 avatar->part[HEAD]->position = mav_matrixMult(avatar->part[HEAD]->position, mat);
 mat = setScalingMatrix(1.3,1.3,1.3);
 avatar->part[RIGHT_UPPER_ARM]->position = mav_matrixMult(avatar->part[RIGHT_UPPER_ARM]->position, mat);
 avatar->part[RIGHT_LOWER_ARM]->position = mav_matrixMult(avatar->part[RIGHT_LOWER_ARM]->position, mat);
 avatar->part[RIGHT_HAND]->position = mav_matrixMult(avatar->part[RIGHT_HAND]->position, mat);
 avatar->part[LEFT_UPPER_ARM]->position = mav_matrixMult(avatar->part[LEFT_UPPER_ARM]->position, mat);
 avatar->part[LEFT_LOWER_ARM]->position = mav_matrixMult(avatar->part[LEFT_LOWER_ARM]->position, mat);
 avatar->part[LEFT_HAND]->position = mav_matrixMult(avatar->part[LEFT_HAND]->position, mat);
 }

/**************************************************************************************************/

/**********************************************/
/* calculate view modifier from head position */
/**********************************************/
void avatarView(MAV_window *w)
 {
 MAV_matrix mat;

 // get back matrix
 mat = backMatrix;

 // eye point relative to torso
 w->vp->trans_eye = mav_vectorMult(mav_vectorMult(*eyePoint, mav_matrixInverse(torsoScale)), mat);

 // get position of head
 mat = mav_matrixMult(mat, avatar->part[NECK]->position);
 mat = mav_matrixMult(mat, avatar->part[NECK]->rotation);

 mat = mav_matrixMult(mat, avatar->part[HEAD]->position);
 mat = mav_matrixMult(mat, avatar->part[HEAD]->rotation);

 w->vp->trans_view = mav_matrixZAxisGet(mat);
 w->vp->trans_up = mav_matrixYAxisGet(mat);
 w->vp->trans_right = mav_vectorScalar(mav_matrixXAxisGet(mat), -1);

 }

/************************************/
/* make view float far above avatar */
/************************************/
void floatingFarView(MAV_window *w)
 {
 MAV_vector pos = mav_matrixXYZGet(avatar->matrix);
 pos.y = pos.y+1;
 pos.z = pos.z+1;

 w->vp->trans_eye = pos;

 // calculate eye view looking down towards avatar
 w->vp->trans_right = MAV_X_VECTOR;
 w->vp->trans_view = mav_vectorRotate(mav_vectorSet(0,0,-1), w->vp->trans_right, MAV_DEG2RAD(-45));
 w->vp->trans_up = mav_vectorCrossProduct(w->vp->trans_right, w->vp->trans_view);
 }

/***************************************************************/
/* make view float near to avatar (and restrict head movement) */
/***************************************************************/
void floatingNearView(MAV_window *w)
 {
 MAV_matrix mat;
 MAV_vector pos;
 MAV_vector offset = mav_vectorSet(0, 1.3, 0);

 // get translation to hips
 mat = mav_matrixMult(MAV_ID_MATRIX, avatar->matrix);
 mat = mav_matrixMult(mat, avatar->vertical);
 mat = mav_matrixMult(mat, avatar->rotation);
 mat = mav_matrixMult(mat, avatar->part[HIPS]->position);
 mat = mav_matrixMult(mat, avatar->part[HIPS]->rotation);

 // shift veritcally up from hip position
 pos = mav_matrixXYZGet(mat);
 pos = mav_vectorAdd(pos, mav_vectorMult(offset, torsoScale));

 w->vp->trans_eye = pos;

 // calculate eye view looking down towards avatar
 w->vp->trans_right = MAV_X_VECTOR;
 w->vp->trans_view = mav_vectorRotate(mav_vectorSet(0,0,-1), w->vp->trans_right, MAV_DEG2RAD(-60));
 w->vp->trans_up = mav_vectorCrossProduct(w->vp->trans_right, w->vp->trans_view);
 }

/**************************************************************************************************/

/****************************/
/* walk navigation function */
/****************************/
//NOTE: can use view direction rather than legPosition, as view always points in direction avatar is
//      standing.  The view modifiers are applied to change view relative to position of head.
void walk(MAV_viewParams *vp, float amount, float ls, float as)
 {
 float xAmount, yAmount, zAmount;
 MAV_vector oldPosition = mav_matrixXYZGet(avatar->matrix);

 // calculate change in position
 xAmount = (avatarDirection.x * amount) * ls;
 yAmount = (avatarDirection.y * amount) * ls;
 zAmount = (avatarDirection.z * amount) * ls;

 // move avatar
 avatar->matrix = mav_matrixXYZSet(avatar->matrix, mav_vectorAdd(oldPosition, mav_vectorSet(xAmount, yAmount, zAmount)));
 }

/***********************************/
/* turn (body) navigation function */
/***********************************/
void turn(MAV_viewParams *vp, float amount, float ls, float as)
 {
 // calculate turn
 float scaledAmount = amount*as;

 // turn avatar
 avatarYaw += MAV_RAD2DEG(scaledAmount);
 }

/*************************************/
/* head movement navigation function */
/*************************************/
void yawHead(MAV_viewParams *vp, float amount, float ls, float as)
 {
 float newYaw = headYaw + MAV_RAD2DEG(amount*as);
 /*if (newYaw>-80.0 && newYaw<80.0)*/ headYaw = newYaw;
 // avatars head turned in rendering loop
 }

void pitchHead(MAV_viewParams *vp, float amount, float ls, float as)
 {
 float newPitch = headPitch + MAV_RAD2DEG(amount*as);
 /*if (newPitch>-85.0 && newPitch<85.0)*/ headPitch = newPitch;
 // avatars head tilted in rendering loop
 }

/**************************************************************************************************/

/*************************************/
/* set initial sensor location info. */
/*************************************/
// BUG WARNING: left hand sensor -> avatar translations are wrong (the hand's upside down).
void initTrackers(int torsoSensor)
 {
 char choice;
 MAV_matrix mat;

 // set sensor-avatar axis translations
 // head
 headSATrans = mav_matrixXAxisSet(MAV_ID_MATRIX, mav_vectorScalar(MAV_X_VECTOR,-1));
 headSATrans = mav_matrixYAxisSet(headSATrans, MAV_Y_VECTOR);
 headSATrans = mav_matrixZAxisSet(headSATrans, mav_vectorScalar(MAV_Z_VECTOR,-1));

 // hands
 leftHandSATrans = mav_matrixXAxisSet(MAV_ID_MATRIX, MAV_X_VECTOR);
 leftHandSATrans = mav_matrixYAxisSet(leftHandSATrans, MAV_Z_VECTOR);
 leftHandSATrans = mav_matrixZAxisSet(leftHandSATrans, mav_vectorScalar(MAV_Y_VECTOR,-1));
 rightHandSATrans = mav_matrixXAxisSet(MAV_ID_MATRIX, MAV_X_VECTOR);
 rightHandSATrans = mav_matrixYAxisSet(rightHandSATrans, MAV_Z_VECTOR);
 rightHandSATrans = mav_matrixZAxisSet(rightHandSATrans, mav_vectorScalar(MAV_Y_VECTOR,-1));

 // back
 if (torsoSensor == BACK)
   {
   torsoSATrans = mav_matrixXAxisSet(MAV_ID_MATRIX, mav_vectorScalar(MAV_X_VECTOR,-1));
   torsoSATrans = mav_matrixYAxisSet(torsoSATrans, mav_vectorScalar(MAV_Z_VECTOR,-1));
   torsoSATrans = mav_matrixZAxisSet(torsoSATrans, mav_vectorScalar(MAV_Y_VECTOR,-1));
   }
 else // sensor strapped on front
   {
   torsoSATrans = mav_matrixXAxisSet(MAV_ID_MATRIX, MAV_X_VECTOR);
   torsoSATrans = mav_matrixYAxisSet(torsoSATrans, mav_vectorScalar(MAV_Z_VECTOR,-1));
   torsoSATrans = mav_matrixZAxisSet(torsoSATrans, MAV_Y_VECTOR);
   }

 // set torso initial position correction
 mat = mav_TDM_pos[MAV_TDM_VEL].matrix; // read torso sensor
 mat = mav_matrixMult(mav_matrixSet(0,0,180,0,0,0), mat); // rotate 180 about world y-axis (as facing polhemus tracker)
 mat = mav_matrixMult(mat, torsoSATrans); // translate to avatar coordinates
 mat = mav_matrixXYZSet(mat, mav_vectorSet(0,0,0)); // remove position components
 torsoCorrection = mav_matrixMult(mav_matrixInverse(MAV_ID_MATRIX), mat); // get difference between initial torso position and upright position
 torsoCorrection = mav_matrixInverse(torsoCorrection); // inverse to get desired correction matrix

 printf("initialise rough sensor offsets? (y or n): ");
 scanf("%s", &choice);
 if (choice != 'n')
   {
   // rough offsets (in avatar part co-ordinate systems)
   eyeOffset = mav_vectorSet(0, -12, 14); // from HMD sensor (in cm)
   handOffset = mav_vectorSet(0, -8, -1); // from hand sensor (3D mouse) (in cm)
   if (torsoSensor == BACK)
     torsoOffset = mav_vectorSet(0, 0, 4); // from back sensor (in cm)
   else
     torsoOffset = mav_vectorSet(0, 13, -13); // from front sensor (in cm)
   // (in sensor co-ordinate system)
   upperArmOffset = mav_vectorSet(-5, 0, 0); // from upper arm sensor (in cm)
   }
 printf("setting torso sensor offset to (%f, %f, %f)\n", torsoOffset.x, torsoOffset.y, torsoOffset.z);
 printf("setting upper arm sensor offset to (%f, %f, %f)\n", upperArmOffset.x, upperArmOffset.y, upperArmOffset.z);
 printf("setting hand sensor offset to (%f, %f, %f)\n", handOffset.x, handOffset.y, handOffset.z);
 printf("setting eye sensor offset to (%f, %f, %f)\n", eyeOffset.x, eyeOffset.y, eyeOffset.z);

 // set sensor scales
 mav_TDMSingleXYZScaleSet(MAV_TDM_HMD, TRACKER_SCALE, TRACKER_SCALE, TRACKER_SCALE);
 mav_TDMSingleOffsetSet(MAV_TDM_HMD, 0);
 mav_TDMSingleDrawScaleSet(MAV_TDM_HMD, 0.01);

 mav_TDMSingleXYZScaleSet(MAV_TDM_RED, TRACKER_SCALE, TRACKER_SCALE, TRACKER_SCALE); // convert from inches to meters
 mav_TDMSingleOffsetSet(MAV_TDM_RED, 0);
 mav_TDMSingleDrawScaleSet(MAV_TDM_RED, 0.01);

 mav_TDMSingleXYZScaleSet(MAV_TDM_BLUE, TRACKER_SCALE, TRACKER_SCALE, TRACKER_SCALE);
 mav_TDMSingleOffsetSet(MAV_TDM_BLUE, 0);
 mav_TDMSingleDrawScaleSet(MAV_TDM_BLUE, 0.01);

 mav_TDMSingleXYZScaleSet(MAV_TDM_VEL, TRACKER_SCALE, TRACKER_SCALE, TRACKER_SCALE);
 mav_TDMSingleOffsetSet(MAV_TDM_VEL, 0);
 mav_TDMSingleDrawScaleSet(MAV_TDM_VEL, 0.01);

 }

/***************/
/* set offsets */
/***************/
void setOffsets(void)
 {
 char choice;
 float x,y,z;

 printf("set torso, arm, hand, or eye offset? (t, a, h, or e): ");
 scanf("%s", &choice);
 switch (choice)
   {
   case 't':
     printf("please enter torso sensor offset x y z (in avatar co-ords): ");
     scanf("%f %f %f", &x, &y, &z);
     printf("setting torso sensor offset to (%f, %f, %f)\n", x,y,z);
     torsoOffset = mav_vectorSet(x,y,z);
     break;
   case 'a':
     printf("please enter upper arm sensor offset x y z (in sensor co-ords): ");
     scanf("%f %f %f", &x, &y, &z);
     printf("setting upper arm sensor offset to (%f, %f, %f)\n", x,y,z);
     upperArmOffset = mav_vectorSet(x,y,z);
     break;
   case 'h':
     printf("please enter hand sensor offset x y z (in avatar co-ords): ");
     scanf("%f %f %f", &x, &y, &z);
     printf("setting hand sensor offset to (%f, %f, %f)\n", x,y,z);
     handOffset = mav_vectorSet(x,y,z);
     break;
   case 'e':
     printf("please enter eye sensor offset x y z (in avatar co-ords): ");
     scanf("%f %f %f", &x, &y, &z);
     printf("setting eye sensor offset to (%f, %f, %f)\n", x,y,z);
     eyeOffset = mav_vectorSet(x,y,z);
     break;
   default:
     printf("not adjusting offset:- must specify which offset to adjust\n");
     break;
   }
 }

/****************************/
/* put head in HMD position */
/****************************/
void trackHead(void *unused)
 {
 MAV_matrix head, back;

 // get HMD position
 head = mav_TDM_pos[MAV_TDM_HMD].matrix; // read sensor
 head = mav_matrixMult(mav_matrixSet(0,0,180,0,0,0), head); // rotate by 180 about world y-axis (as facing polhemus tracker)
 head = mav_matrixMult(head, headSATrans); // convert from sensor to avatar axis
 head = mav_matrixMult(head, mav_matrixXYZSet(MAV_ID_MATRIX, eyeOffset)); // apply offset to eye

 // get orientation of HMD sensor relative to back
 back = mav_matrixMult(MAV_ID_MATRIX, avatar->part[HIPS]->position);
 back = mav_matrixMult(back, avatar->part[HIPS]->rotation);
 back = mav_matrixMult(back, avatar->part[LOWER_TORSO]->position);
 back = mav_matrixMult(back, avatar->part[LOWER_TORSO]->rotation);
 back = mav_matrixMult(back, avatar->part[UPPER_TORSO]->position);
 back = mav_matrixMult(back, avatar->part[UPPER_TORSO]->rotation);

 // get head relative to back (upper torso)
 head = mav_matrixMult(mav_matrixInverse(back), head);
 eyePointTracked = mav_matrixXYZGet(head); // set tracked eye point
 mav_matrixRPYGet(head, &headRoll, &headPitch, &headYaw);  // roll, pitch, and yaw of HMD (relative to back)
 }

/*************************************************/
/* track head, back and hands (relative to back) */
/*************************************************/
// scale mouse position from inches into meters and then apply some offsets
void trackUpperBody(void *unused)
 {
 MAV_vector backPt, l, r;
 MAV_matrix head, torso, left, right;

 // get HMD matrix
 head = mav_TDM_pos[MAV_TDM_HMD].matrix; // read sensor
 head = mav_matrixMult(mav_matrixSet(0,0,180,0,0,0), head); // rotate by 180 about world y-axis (as facing polhemus tracker)
 head = mav_matrixMult(head, headSATrans); // convert from sensor to avatar axis
 head = mav_matrixMult(head, mav_matrixXYZSet(MAV_ID_MATRIX, eyeOffset)); // apply offset to eye

 // get velvet sensor (back) matrix
 torso = mav_TDM_pos[MAV_TDM_VEL].matrix; // read sensor
 torso = mav_matrixMult(mav_matrixSet(0,0,180,0,0,0), torso);
 torso = mav_matrixMult(torso, torsoSATrans); // convert from sensor to avatar axis
 torso = mav_matrixMult(torso, torsoCorrection); // apply correction
 torso = mav_matrixMult(torso, mav_matrixXYZSet(MAV_ID_MATRIX, torsoOffset)); // apply offset

 // set torso rotation
 mav_matrixRPYGet(torso, &torsoRoll, &torsoPitch, &torsoYaw);  // roll, pitch, and yaw torso

 // head relative to torso (torso sensor)
 head = mav_matrixMult(mav_matrixInverse(torso), head);
 eyePointTracked = mav_vectorScalar(mav_matrixXYZGet(head),TRACKER_SCALE); // set tracked eye point
 mav_matrixRPYGet(head, &headRoll, &headPitch, &headYaw); // set rotation

 // get matrix of hand TDMs
 left = mav_TDM_pos[MAV_TDM_RED].matrix; // read sensor
 left = mav_matrixMult(mav_matrixSet(0,0,180,0,0,0), left);
 right = mav_TDM_pos[MAV_TDM_BLUE].matrix;
 right = mav_matrixMult(mav_matrixSet(0,0,180,0,0,0), right);
 left = mav_matrixMult(left, leftHandSATrans); // convert from sensor to avatar axis
 right = mav_matrixMult(right, rightHandSATrans);
 left = mav_matrixMult(left, mav_matrixXYZSet(MAV_ID_MATRIX, handOffset));  // apply offsets
 right = mav_matrixMult(right, mav_matrixXYZSet(MAV_ID_MATRIX, handOffset));

 // get hand matrix relative to torso matrix
 left = mav_matrixMult(mav_matrixInverse(torso), left);
 right = mav_matrixMult(mav_matrixInverse(torso), right);

 leftHand = mav_vectorScalar(mav_matrixXYZGet(left), TRACKER_SCALE);
 rightHand = mav_vectorScalar(mav_matrixXYZGet(right), TRACKER_SCALE);

 // get hand rotation relative to torso
 mav_matrixRPYGet(left, &leftHandRoll, &leftHandPitch, &leftHandYaw); // roll, pitch, and yaw hands
 mav_matrixRPYGet(right, &rightHandRoll, &rightHandPitch, &rightHandYaw);

 }

/*********************/
/* track right elbow */
/*********************/
void trackElbow(void *unused)
 {
 MAV_vector normal, vec;
// MAV_BB bb;
 MAV_matrix mat, elbow, torso;

 // get elbow position
 elbow = mav_TDM_pos[MAV_TDM_RED].matrix;
 elbow = mav_matrixMult(mav_matrixSet(0,0,180,0,0,0), elbow);
 // apply offset
 elbow = mav_matrixMult(elbow, mav_matrixXYZSet(MAV_ID_MATRIX, upperArmOffset));

 // get torso matrix
 torso = mav_TDM_pos[MAV_TDM_VEL].matrix;
 torso = mav_matrixMult(mav_matrixSet(0,0,180,0,0,0), torso);
 torso = mav_matrixMult(torso, torsoSATrans); // convert from sensor to avatar axis
 torso = mav_matrixMult(torso, mav_matrixXYZSet(MAV_ID_MATRIX, torsoOffset)); // apply offset

 vec = mav_matrixXYZGet(elbow);
 vec = mav_vectorScalar(mav_vectorMult(vec, mav_matrixInverse(torso)), TRACKER_SCALE); // get elbow relative to torso

 // move along torso to shoulder
 vec.x += shoulderWidth; // would subtract if tracking the left elbow

 // set roll, yaw and pitch of right arm relative to shoulder
 rightArmRollTracked = 0;
 getPY(vec, &rightArmPitchTracked, &rightArmYawTracked);

 }

/**************************************************************************************************/

/*************/
/* load a NN */
/*************/
int loadNetwork(char *filename)
 {
 if (net != NULL) // if a net already exists, free the memory it occupies
   {
   freeNet(net);
   }

 printf("\nloading network\n");

 net = loadNet(filename);
 if (net==NULL)
   {
   printf("could not open '%s'\n",filename);
   return 0;
   }

 return 1;
 }

/****************************/
/* calculate elbow position */
/****************************/
void predictElbows(void *unused)
 {
 float rPitch, rYaw, r, p, y;
 MAV_matrix mat;
 MAV_vector vec, xAxis, yAxis, zAxis;

 if (net==NULL) return; // do nothing if a NN is not loaded

 if (net->input_n==6 && net->output_n==2)
   {

   // present right hand data to neural network     
   predict(net, shoulderWidth, upperArmLength, lowerArmLength, rightHand, rightHandRoll, rightHandPitch, rightHandYaw, &rYaw, &rPitch); // right hand
      
   // set right arm globals
   rightArmRollNetwork = 0;
   rightArmPitchNetwork = rPitch;
   rightArmYawNetwork = rYaw;

   // get left hand orientation as a matrix
   mat = mav_matrixRPYSet(MAV_ID_MATRIX, leftHandRoll, leftHandPitch, leftHandYaw);
   
   // reflect the axes in the YZ plane
   xAxis = mav_matrixXAxisGet(mat);
   xAxis.x = -xAxis.x;
   yAxis = mav_matrixYAxisGet(mat);
   yAxis.x = -yAxis.x;
   zAxis = mav_matrixZAxisGet(mat);
   zAxis.x = -zAxis.x;
   
   // ensure the axes remain orthogonal
   zAxis = mav_vectorCrossProduct(xAxis, yAxis); // the axes should already be unit vectors
   xAxis = mav_vectorCrossProduct(yAxis, zAxis);
   yAxis = mav_vectorCrossProduct(zAxis, xAxis);
   
   // re-form the matrix
   mat = mav_matrixXAxisSet(MAV_ID_MATRIX, mav_vectorNormalise(xAxis));
   mat = mav_matrixYAxisSet(mat, mav_vectorNormalise(yAxis));
   mat = mav_matrixZAxisSet(mat, mav_vectorNormalise(zAxis));
   
   // get the RPY
   mav_matrixRPYGet(mat, &r, &p, &y); 
    
   // flip left hand position
   vec = flipHand(leftHand);
   
   // present left hand data to neural network
   predict(net, shoulderWidth, upperArmLength, lowerArmLength, vec, r, p, y, &rYaw, &rPitch); // left hand
   
   // flip output so it represents left hand
   flipPY(rPitch, rYaw, &p, &y);

   // set left arm globals
   leftArmRollNetwork = 0;
   leftArmPitchNetwork = p;
   leftArmYawNetwork = y;
   
   }
 else return; // do nothing if the network has incorrect number of inputs or outputs

 }

/**************************************************************************************************/

void testHands(void *unused)
 {
 printf("\nleft local x=%f y=%f z=%f\n", leftHand.x, leftHand.y, leftHand.z);
 printf("right local x=%f y=%f z=%f\n", rightHand.x, rightHand.y, rightHand.z);
 printf("left avatar x=%f y=%f z=%f\n", avatar->left_hand.x, avatar->left_hand.y, avatar->left_hand.z);
 printf("right avatar x=%f y=%f z=%f\n", avatar->right_hand.x, avatar->right_hand.y, avatar->right_hand.z);
 }

/***************************/
/* body movement functions */
/***************************/

// walk movements
void moveForwards(void *unused){ walk(mav_win_current->vp, 1, 0.5/mav_fps, 1); }

void moveBackwards(void *unused){ walk(mav_win_current->vp, -1, 0.5/mav_fps, 1); }

void turnRight(void *unused){ turn(mav_win_current->vp, -1, 1, 0.5/mav_fps); }

void turnLeft(void *unused){ turn(mav_win_current->vp, 1, 1, 0.5/mav_fps); }

// arm movements
void leftHandFurther(void *unused){ leftHand.z += 0.5/mav_fps; }

void leftHandCloser(void *unused){ leftHand.z -= 0.5/mav_fps; }

void leftHandUp(void *unused){ leftHand.y += 0.5/mav_fps; }

void leftHandDown(void *unused){ leftHand.y -= 0.5/mav_fps; }

void leftHandRight(void *unused){ leftHand.x -= 0.5/mav_fps; }

void leftHandLeft(void *unused){ leftHand.x += 0.5/mav_fps; }

void rightHandFurther(void *unused){ rightHand.z += 0.5/mav_fps; }

void rightHandCloser(void *unused){ rightHand.z -= 0.5/mav_fps; }

void rightHandUp(void *unused){ rightHand.y += 0.5/mav_fps; }

void rightHandDown(void *unused){ rightHand.y -= 0.5/mav_fps; }

void rightHandRight(void *unused){ rightHand.x -= 0.5/mav_fps; }

void rightHandLeft(void *unused){ rightHand.x += 0.5/mav_fps; }

// hand movements
void leftHandRollLeft(void *unused){ leftHandRoll -= MAV_RAD2DEG(0.5/mav_fps); }

void leftHandRollRight(void *unused){ leftHandRoll += MAV_RAD2DEG(0.5/mav_fps); }

void leftHandPitchDown(void *unused){ leftHandPitch += MAV_RAD2DEG(0.5/mav_fps); }

void leftHandPitchUp(void *unused){ leftHandPitch -= MAV_RAD2DEG(0.5/mav_fps); }

void leftHandYawLeft(void *unused){ leftHandYaw += MAV_RAD2DEG(0.5/mav_fps); }

void leftHandYawRight(void *unused){ leftHandYaw -= MAV_RAD2DEG(0.5/mav_fps); }

void rightHandRollLeft(void *unused){ rightHandRoll -= MAV_RAD2DEG(0.5/mav_fps); }

void rightHandRollRight(void *unused){ rightHandRoll += MAV_RAD2DEG(0.5/mav_fps); }

void rightHandPitchDown(void *unused){ rightHandPitch -= MAV_RAD2DEG(0.5/mav_fps); }

void rightHandPitchUp(void *unused){ rightHandPitch += MAV_RAD2DEG(0.5/mav_fps); }

void rightHandYawLeft(void *unused){ rightHandYaw -= MAV_RAD2DEG(0.5/mav_fps); }

void rightHandYawRight(void *unused){ rightHandYaw += MAV_RAD2DEG(0.5/mav_fps); }

// elbow movements
void leftArmPitchUp(void *unused){ leftArmPitchKeyboard -= MAV_RAD2DEG(0.5/mav_fps); }

void leftArmPitchDown(void *unused){ leftArmPitchKeyboard += MAV_RAD2DEG(0.5/mav_fps); }

void leftArmYawRight(void *unused){ leftArmYawKeyboard -= MAV_RAD2DEG(0.5/mav_fps); }

void leftArmYawLeft(void *unused){ leftArmYawKeyboard += MAV_RAD2DEG(0.5/mav_fps); }

void rightArmPitchUp(void *unused){ rightArmPitchKeyboard -= MAV_RAD2DEG(0.5/mav_fps); }

void rightArmPitchDown(void *unused){ rightArmPitchKeyboard += MAV_RAD2DEG(0.5/mav_fps); }

void rightArmYawRight(void *unused){ rightArmYawKeyboard -= MAV_RAD2DEG(0.5/mav_fps); }

void rightArmYawLeft(void *unused){ rightArmYawKeyboard += MAV_RAD2DEG(0.5/mav_fps); }

// head movements
void fixHeadLookingForward(void *unused)
 {
 headRoll = 0;
 headPitch = 0;
 headYaw = 0;
 }

void fixHeadLookingDown(void *unused)
 {
 headRoll = 0;
 headPitch = 30;
 headYaw = 0;
 }

void headLeft(void *unused){ yawHead(mav_win_current->vp, 1, 1, 0.5/mav_fps); }

void headRight(void *unused){ yawHead(mav_win_current->vp, -1, 1, 0.5/mav_fps); }

void headUp(void *unused){ pitchHead(mav_win_current->vp, -1, 1, 0.5/mav_fps); }

void headDown(void *unused){ pitchHead(mav_win_current->vp, 1, 1, 0.5/mav_fps); }

// torso movements
void torsoRollRight(void *unused)
 {
 float newRoll = torsoRoll + MAV_RAD2DEG(0.5/mav_fps);
 /*if (newRoll>-40.0 && newRoll<40.0)*/ torsoRoll = newRoll;
 }

void torsoRollLeft(void *unused)
 {
 float newRoll = torsoRoll - MAV_RAD2DEG(0.5/mav_fps);
 /*if (newRoll>-40.0 && newRoll<40.0)*/ torsoRoll = newRoll;
 }

void torsoPitchBack(void *unused)
 {
 float newPitch = torsoPitch - MAV_RAD2DEG(0.5/mav_fps);
 /*if (newPitch>-40.0 && newPitch<85.0)*/ torsoPitch = newPitch;
 }

void torsoPitchForward(void *unused)
 {
 float newPitch = torsoPitch + MAV_RAD2DEG(0.5/mav_fps);
 /*if (newPitch>-40.0 && newPitch<85.0)*/ torsoPitch = newPitch;
 }

void torsoYawRight(void *unused)
 {
 float newYaw = torsoYaw - MAV_RAD2DEG(0.5/mav_fps);
 /*if (newYaw>-80.0 && newYaw<80.0)*/ torsoYaw = newYaw;
 }

void torsoYawLeft(void *unused)
 {
 float newYaw = torsoYaw + MAV_RAD2DEG(0.5/mav_fps);
 /*if (newYaw>-80.0 && newYaw<80.0)*/ torsoYaw = newYaw;
 }

/**************************************************************************************************/

/*******************/
/* position avatar */
/*******************/

void positionAvatar(void *unused)
 {
 MAV_matrix mat, head, back;
 MAV_vector vec;
 float r,p,y;

 // set yaw of avatar
 avatar->matrix = mav_matrixRPYSet(avatar->matrix, 0, 0, avatarYaw);

 // set avatar direction
 mat = mav_matrixMult(MAV_ID_MATRIX, avatar->matrix);
 mat = mav_matrixMult(mat, avatar->vertical);
 mat = mav_matrixMult(mat, avatar->rotation);
 avatarDirection = mav_matrixZAxisGet(mat);

 // set torso rotation
 avatar->part[LOWER_TORSO]->rotation = mav_matrixRPYSet(avatar->part[LOWER_TORSO]->rotation, torsoRoll, torsoPitch, torsoYaw);

 // set head rotation (positive yaw means yaw to left, and positive pitch means pitch down)
 avatar->part[HEAD]->rotation = mav_matrixRPYSet(avatar->part[HEAD]->rotation, headRoll, headPitch, headYaw);

 // get translation from upper torso to head
 head = mav_matrixMult(MAV_ID_MATRIX, avatar->part[NECK]->position);
 head = mav_matrixMult(head, avatar->part[NECK]->rotation);
 head = mav_matrixMult(head, avatar->part[HEAD]->position);
 head = mav_matrixMult(head, avatar->part[HEAD]->rotation);

 // set eye point relative to upper torso
 eyePointAvatar = mav_vectorMult(mav_vectorSet(0,0.1,0.011), head);

 // hand position is relative to back, get relative to world for the existing mechanical arm positioning
 back = mav_matrixMult(mat,avatar->part[HIPS]->position);
 back = mav_matrixMult(back,avatar->part[HIPS]->rotation);

 back = mav_matrixMult(back,avatar->part[LOWER_TORSO]->position);
 back = mav_matrixMult(back,avatar->part[LOWER_TORSO]->rotation);

 back = mav_matrixMult(back,avatar->part[UPPER_TORSO]->position);
 back = mav_matrixMult(back,avatar->part[UPPER_TORSO]->rotation);

 backMatrix = back;

 // set world hand matrix
 rightHandMatrix = mav_matrixXYZSet(MAV_ID_MATRIX, mav_vectorMult(rightHand, mav_matrixInverse(torsoScale)));
 rightHandMatrix = mav_matrixRPYSet(rightHandMatrix, rightHandRoll, rightHandPitch, rightHandYaw);
 leftHandMatrix = mav_matrixXYZSet(MAV_ID_MATRIX, mav_vectorMult(leftHand, mav_matrixInverse(torsoScale)));
 leftHandMatrix = mav_matrixRPYSet(leftHandMatrix, leftHandRoll, leftHandPitch, leftHandYaw);

 // hands relative to back 
 rightHandMatrix = mav_matrixMult(back, rightHandMatrix);
 leftHandMatrix = mav_matrixMult(back, leftHandMatrix);

 // set world hand positions
 avatar->right_hand = mav_matrixXYZGet(rightHandMatrix);
 avatar->left_hand = mav_matrixXYZGet(leftHandMatrix);

 // set arms
 if (holdingLeft)
   setArm(LEFT_ARM);
 if (holdingRight)
   setArm(RIGHT_ARM);

 }

/********************/
/* set a single arm */
/********************/
void setArm(int arm)
 {
 MAV_vector normal, wrist;
 MAV_matrix handMat;
 MAV_BB bb;
 float roll, pitch, yaw, handRoll, handPitch, handYaw;
 float totalArmLength;
 int laser = 0;
 int c, ua, la, h;
 float distance, cosAngle, angle, correctedAngle;
 MAV_matrix mat, lowerArmMat;
 MAV_vector handPos;

 // set left or right arm
 if (arm == LEFT_ARM)
   {
   c = LEFT_CLAVICLE;
   ua = LEFT_UPPER_ARM;
   la = LEFT_LOWER_ARM;
   h = LEFT_HAND;
   hand = leftHand;
   roll = *leftArmRoll;
   pitch = *leftArmPitch;
   yaw = *leftArmYaw;
   handRoll = leftHandRoll;
   handPitch = leftHandPitch;
   handYaw = leftHandYaw;
   }
 else if (arm == RIGHT_ARM)
   {
   c = RIGHT_CLAVICLE;
   ua = RIGHT_UPPER_ARM;
   la = RIGHT_LOWER_ARM;
   h = RIGHT_HAND;
   hand = rightHand;
   roll = *rightArmRoll;
   pitch = *rightArmPitch;
   yaw = *rightArmYaw;
   handRoll = rightHandRoll;
   handPitch = rightHandPitch;
   handYaw = rightHandYaw;
   }
 else return;

 // set total arm length
 totalArmLength = upperArmLength+lowerArmLength;

 // set elbow position relative to shoulder
 predictedUpperArm = mav_vectorMult(MAV_Y_VECTOR, mav_matrixRPYSet(MAV_ID_MATRIX, roll, pitch, yaw));
 // since roll=0 this should be equivalent to...
 // vec = mav_vectorRotate(MAV_Y_VECTOR, MAV_X_VECTOR, MAV_DEG2RAD(armPitch));
 // vec = mav_vectorRotate(vec, MAV_Y_VECTOR, MAV_DEG2RAD(armYaw));
 // or
 // vec = mav_vectorRotate(MAV_X_VECTOR, MAV_Y_VECTOR, MAV_DEG2RAD(armYaw));
 // vec = mav_vectorRotate(MAV_Y_VECTOR, vec, MAV_DEG2RAD(armPitch));
 // ensure correct length
 predictedUpperArm = mav_vectorScalar(predictedUpperArm, -upperArmLength/mav_vectorMag(predictedUpperArm));

 // get hand relative to shoulder rather than back
 if (arm == LEFT_ARM) hand.x -= shoulderWidth;
 else hand.x += shoulderWidth;

 // get distance of hand from shoulder
 distance = mav_vectorMag(hand);

 // constrain distance from shoulder to hand (used for correcting angle)
 if ( distance>totalArmLength ) // if distance from shoulder to hand is greater than total arm length
   {
   distance = totalArmLength; // set distance as total arm length
   hand = mav_vectorScalar(hand, totalArmLength/mav_vectorMag(hand)); // scale shoulder-to-hand vector to total arm length
   laser = 1; // turn laser on
   }
 else
   {
//   distance = mav_vectorMag(hand); // set as distance from shoulder to hand
   laser = 0;
   }

 // calculate correct angle
 cosAngle = ((upperArmLength*upperArmLength)+(distance*distance)-(lowerArmLength*lowerArmLength)) / (2.0*upperArmLength*distance);
// if (cosAngle<0) correctedAngle = acos(-cosAngle);
// else correctedAngle = acos(cosAngle);
 correctedAngle = acos(cosAngle);

 // get elbow relative to back rather than shoulder
 if (arm == LEFT_ARM) predictedUpperArm.x += shoulderWidth;
 else predictedUpperArm.x -= shoulderWidth;

 // get hand relative to back (from shoulder)
 if (arm == LEFT_ARM) hand.x += shoulderWidth;
 else hand.x -= shoulderWidth;

 // get upper arm and hand relative to clavicle
 mat = torsoScale; // take into account torso scale (other scales are incorporated in the parts position matrix)
 mat = mav_matrixMult(mat, avatar->part[c]->position);
 mat = mav_matrixMult(mat, avatar->part[c]->rotation);
 mat = mav_matrixMult(mat, avatar->part[ua]->position);

 // get upper arm and hand relative to end of clavicle rather than back
 predictedUpperArm = mav_vectorMult(predictedUpperArm, mav_matrixInverse(mat));
 hand = mav_vectorMult(hand, mav_matrixInverse(mat));

 // calculate normal and angle between upper arm and shoulderToHand vector
 normal = mav_vectorCrossProduct(mav_vectorNormalise(predictedUpperArm), mav_vectorNormalise(hand));
 cosAngle = mav_vectorDotProduct(mav_vectorNormalise(predictedUpperArm), mav_vectorNormalise(hand));
// if (cosAngle<0) angle = acos(-cosAngle);
// else angle = acos(cosAngle);
 angle = acos(cosAngle);

 // correct upper arm
 upperArm = mav_vectorRotate(predictedUpperArm, mav_vectorNormalise(normal), angle-correctedAngle); // rotate upper arm by difference between angles about normal.

 // set upper arm rotation
 getPY(upperArm, &pitch, &yaw);
 avatar->part[ua]->rotation = mav_matrixRPYSet(MAV_ID_MATRIX, 0, pitch, yaw);

 // get lower arm vector (now relative to elbow)
 lowerArm = mav_vectorSub(hand, upperArm);
 lowerArm = mav_vectorMult(lowerArm, mav_matrixInverse(avatar->part[ua]->rotation)); // get relative to upper arm rotation

 // get hand rotation relative to elbow rotation
 mat = mav_matrixMult(MAV_ID_MATRIX, avatar->part[c]->rotation);
 mat = mav_matrixMult(mat, avatar->part[ua]->rotation);
 handMat = mav_matrixMult(mav_matrixInverse(mat), mav_matrixRPYSet(MAV_ID_MATRIX, handRoll, handPitch, handYaw));

 // calculate xyz vectors of lower arm from lowerArm vector and hand rotation
 yAxis = mav_vectorNormalise( mav_vectorScalar(lowerArm,-1) );
 xAxis = mav_vectorCrossProduct(mav_vectorNormalise(mav_matrixZAxisGet(handMat)), mav_vectorNormalise(yAxis));
 xAxis = mav_vectorNormalise(xAxis);
 zAxis = mav_vectorNormalise(mav_vectorCrossProduct(xAxis, yAxis));

 // form lower arm rotation matrix
 lowerArmMat = mav_matrixZAxisSet(MAV_ID_MATRIX, zAxis);
 lowerArmMat = mav_matrixYAxisSet(lowerArmMat, yAxis);
 lowerArmMat = mav_matrixXAxisSet(lowerArmMat, xAxis);

 // set lower arm rotation
 avatar->part[la]->rotation = lowerArmMat;

 // set hand rotation
 mat = mav_matrixMult(mat, avatar->part[la]->rotation); // get rotation relative to lower arm
 handMat = mav_matrixMult(mav_matrixInverse(mat), mav_matrixRPYSet(MAV_ID_MATRIX, handRoll, handPitch, handYaw));
 avatar->part[h]->rotation = mav_matrixXYZSet(handMat, mav_vectorSet(0,0,0));

 // if pointing beyond reach draw a laser-pointer
 if (laser)
   {
   // calculate position of wrist
   mat = backMatrix;
   mat = mav_matrixMult(mat, avatar->part[ua]->position);
   mat = mav_matrixMult(mat, avatar->part[ua]->rotation);
   mat = mav_matrixMult(mat, avatar->part[la]->position);
   mat = mav_matrixMult(mat, avatar->part[la]->rotation);
   wrist = mav_matrixXYZGet(mat);

   // draw laser-pointer
   mav_surfaceParamsUse(avatar->laser_sp);

   mav_gfxLineBegin();
    mav_gfxVertex(wrist);
    if (arm == LEFT_ARM) mav_gfxVertex(avatar->left_hand);
    else mav_gfxVertex(avatar->right_hand);
   mav_gfxLineEnd();
   }

 }

/**************************************************************************************************/


