/*
 GNU Maverik - a system for managing display and interaction in 
               Virtual Environment applications.
 Copyright (C) 1999-2001 Advanced Interfaces Group

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA


 The authors can be contacted via:
 www   - http://aig.cs.man.ac.uk
 email - maverik@aig.cs.man.ac.uk
 mail  - Advanced Interfaces Group, Room 2.94, Computer Science Building, 
         University of Manchester, Manchester, M13 9PL, UK
*/


#include "maverik.h"
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>

#include "mav_tdm.h"
#include "avatarlib.h"
#include "iolib.h"

#define PAWN   0
#define ROOK   1
#define KNIGHT 2
#define BISHOP 3
#define QUEEN  4
#define KING   5
#define WHITE  0
#define BLACK  1

int fd[4];
int board[64];
MAV_object *piece[2][6];
int whosgo=1;
int moving=0;
int movefrom=0;
int moveto=0;
MAV_vector movedir;

/**************************************************************************************************/

/********************/
/* global variables */
/********************/
//MAV_surfaceParams *sp[4];
MAV_matrix *objectMatrix; // set in the keyboard callback, when "p" is pressed
float objectDistance; // used to store distance from object/mouse intersection to the eye point

// environment
MAV_rectangle rect;
//MAV_box box;
//MAV_cylinder cyl;
MAV_composite comp;

int viewAtHead = 1; // specify whether view at avatars head, or looking at avatar
int elbowPositioning = 0; // specify mechanical, keyboard yaw and pitch, or neural network prediction of elbow position
int tracking = 0; // whether hands are to track the 3D mice
float eery = 0; // exorcist variable

int testingHands = 0;
int training = 0; // indicates whether gathering training data or not

// application specific data structure
typedef struct
 {
 int number;
 } myStruct;

/**************************************************************************************************/

/***********************/
/* function prototypes */
/***********************/

void pick(void *unused);

void exorcist(void *unused);

int boxMouseEvent(MAV_object *obj, MAV_mouseEvent *mouseEvent);
int compositeMouseEvent(MAV_object *obj, MAV_mouseEvent *mouseEvent);
int tdmEventCallback(MAV_object *obj, MAV_TDMEvent *tdmEvent);

void helpMessage(void *unused);
int keyEventCallback(MAV_object *obj, MAV_keyboardEvent *keyEvent);

void showLines(void);

//void renderFrame(MAV_SMS *sms1, MAV_SMS *sms2);
void renderFrame(MAV_SMS *sms);
int main(int argc, char *argv[]);

/**************************************************************************************************/

/* Routine to draw the board */

void drawboard(void)
{
  MAV_matrix *m;
  MAV_vector p;
  MAV_surfaceParams sp;
  int i, x, y, col=1;

  for (i=0; i<64; i++) {

    /* Calculate the colour of the square */
    if (i%8) col=!col;
      
    sp.mode= MAV_COLOUR;
    if (moving && (i==moveto || i==movefrom)) 
    {
      sp.colour= MAV_COLOUR_RED;
    } 
    else if (col)
    {
      sp.colour= MAV_COLOUR_BLACK;
    }
    else
    {
      sp.colour= MAV_COLOUR_WHITE;
    }
    mav_surfaceParamsUse(&sp);

    /* Draw the square */
    x=i%8;
    y=i/8;

    p.x=x-0.5;
    p.y=y-0.5;
    p.z=0;
    mav_gfxPolygonBegin();
    mav_gfxVertex(p);
    
    p.x+=1.0;
    mav_gfxVertex(p);
    
    p.y+=1.0;
    mav_gfxVertex(p);

    p.x-=1.0;
    mav_gfxVertex(p);
    mav_gfxPolygonEnd();

    /* Does the square contain a piece */
    if (board[i]!=-1) {
      /* calculate which piece and its colour */
      int pc= board[i]&7;
      int co= (board[i]>>4)&1;

      /* get a pointer to the matrix of that piece */
      mav_callbackGetMatrixExec(mav_win_current, piece[co][pc], &m);

      /* set matrix to correct position */
      if (moving && i==movefrom) 
      {
	*m= mav_matrixSet(0,90,0, x+movedir.x*moving, y+movedir.y*moving, 0);
      }
      else
      {
	*m= mav_matrixSet(0,90,0, x, y, 0);
      }
      
      /* render the piece */
      mav_callbackDrawExec(mav_win_current, piece[co][pc], NULL);
    }
  }

  /* Display status of game */
  if (moving) 
  {
    mav_stringDisplay(mav_win_all, "moving", MAV_COLOUR_RED, 0, -0.9, 0.9);
  }
  else if (whosgo) 
  {
    mav_stringDisplay(mav_win_all, "player thinking...", MAV_COLOUR_RED, 0, -0.9, 0.9);
  }
  else
  {
    mav_stringDisplay(mav_win_all, "computer thinking...", MAV_COLOUR_RED, 0, -0.9, 0.9);
  }

  /* Pieces are moved over 5 frames */
  if (moving) {
    moving++;

    if (moving==6) {
    /* movement over, update board */
      moving=0;
      board[moveto]= board[movefrom];
      board[movefrom]= -1;
    }
  }
}



/* Routine to convert chess style position (e.g "h3") into array index */

int sq(char *s)
{
  return ((s[0]-'a')+(s[1]-'1')*8);
}



/* Routine to make the move give in string s */

void movePiece(char *s)
{
  if (strstr(s, "o-o")) 
  {
    if (whosgo) 
    {
      /* player castling */
      board[sq("g1")]= board[sq("e1")];
      board[sq("e1")]= -1;

      board[sq("f1")]= board[sq("h1")]; 
      board[sq("h1")]= -1;
    }
    else
    {
      /* computer castling */
      board[sq("g8")]= board[sq("e8")];
      board[sq("e8")]= -1;

      board[sq("f8")]= board[sq("h8")]; 
      board[sq("h8")]= -1;
    }
  }
  else
  {
    /* calculate the movement of the piece */
    moving= 1;
    movefrom= sq(&s[0]);
    moveto= sq(&s[2]);
    movedir.x= (moveto%8)-(movefrom%8);
    movedir.y= (moveto/8)-(movefrom/8);
    movedir.z= 0.0;
    movedir= mav_vectorScalar(movedir, 0.2); /* complete movement in 5 steps */
  }

  /* toggle whos go */
  whosgo=!whosgo;
}



/* Routine to check for events from GNU chess */

int checkGNUchess(void)
{
  char buf[100];
  fd_set fdset;
  struct timeval td;
  int i;

  for (i=0; i<100; i++) buf[i]=0;
  
  /* Check if any data is waiting to be read */

  FD_ZERO(&fdset);
  FD_SET(fd[0], &fdset);
  
  td.tv_sec=0;
  td.tv_usec=0;

  if (select(100, &fdset, NULL, NULL, &td) && !moving) {

  /* read what GNU chess has to say */
    read(fd[0], buf, 100);

  /* computers move? */
    if (strstr(buf, "is:")) movePiece(strstr(buf, "is:")+4);

  /* players move? */
    if (strstr(buf, "Hint:")) {
      write(fd[3], strstr(buf, "Hint:")+6, 5);
      movePiece(strstr(buf, "Hint:")+6);
    }
  }

  return 1;
}



/* Keyboard event callback routine */

int keyb(MAV_object *o, MAV_keyboardEvent *ke)
{
  if (ke->movement==MAV_PRESSED) {
    
    switch (ke->key) {
    case 'q': /* quit */
      write(fd[3], "exit\n", 5); /* quit gnuchess */
      exit(1);
    default:  /* player makes a move (does this by sending "hint" to gnuchess) */
      if (whosgo && !moving) 
      {
	write(fd[3], "hint\n", 5);
      }
      else
      {
	printf("wait your turn\n");
      }
    }
  }

  return 1;
}


/**************************************************************************************************/

/*********************/
/* pick up an object */
/*********************/
void pick(void *unused)
 {
 MAV_vector position;

 // calculate the position of a point a distance objectDistance away from the eye along
 // the normalized vector mav_mouse_dir (defined by the eye point and the mouse's projection
 // onto the near clip plane).
 // objectDistance is set in the keyboard callback (when "p" is pressed).
 position = mav_vectorAdd(mav_win_current->vp->eye, mav_vectorScalar(mav_mouse_dir, objectDistance));

 // set the objects matrix to this position
 *objectMatrix = mav_matrixXYZSet(*objectMatrix, position);
 }

/**************************************************************************************************/

/************/
/* exorcist */
/************/
void exorcist(void *unused)
 {
 if ((tracking!=1) & (headYaw < (eery+359))) // rotate head (and keep vertical)
   {
   headYaw += 10;
   }
 else // stop rotating head, and reset previous head position
   {
   headYaw = eery;
   mav_frameFn1Rmv(exorcist, NULL);
   exit(1);
   }
 }

/**************************************************************************************************/

/**********************/
/* TDM event callback */
/**********************/

int tdmEventCallback(MAV_object *obj, MAV_TDMEvent *tdmEvent)
 {
 if (tdmEvent->tracker == MAV_TDM_BLUE)
   {
   // start/stop gathering training data if middle button of blue hand sensor is pressed
   if (tdmEvent->button == MAV_TDM_MIDDLE)
     {
     if (tdmEvent->movement == MAV_PRESSED)
       {
       training = !training;
       if (training)
         {
         if (openTrainingFiles())
           mav_frameFn2Add(writeTrainingData, NULL); // added to phase 2 of the rendering loop (after avatars positions have been set)
         else
           {
           closeTrainingFiles();
           training = !training;
           }
         }
       else
         {
         closeTrainingFiles();
         mav_frameFn2Rmv(writeTrainingData, NULL);
         }
       }
     }
   // move forwards if top button of blue hand sensor is pressed
   else if (tdmEvent->button == MAV_TDM_TOP)
     {
     if (tdmEvent->movement == MAV_PRESSED) mav_frameFn1Add(moveForwards, NULL);
     else mav_frameFn1Rmv(moveForwards, NULL);
     }
   // move backwards if bottom button of blue hand sensor is pressed
   else if (tdmEvent->button == MAV_TDM_BOTTOM)
     {
     if (tdmEvent->movement == MAV_PRESSED) mav_frameFn1Add(moveBackwards, NULL);
     else mav_frameFn1Rmv(moveBackwards, NULL);
     }
   }
 else if (tdmEvent->tracker == MAV_TDM_RED)
   {
   // initialise sensors if middle button of red hand sensor is pressed
   if (tdmEvent->button == MAV_TDM_MIDDLE)
     {
     if (tdmEvent->movement == MAV_PRESSED) initTrackers();
     }
   // turn right if top button of red hand sensor is pressed
   else if (tdmEvent->button == MAV_TDM_TOP)
     {
     if (tdmEvent->movement == MAV_PRESSED) mav_frameFn1Add(turnRight, NULL);
     else mav_frameFn1Rmv(turnRight, NULL);
     }
   // turn left if bottom button of red hand sensor is pressed
   else if (tdmEvent->button == MAV_TDM_BOTTOM)
     {
     if (tdmEvent->movement == MAV_PRESSED) mav_frameFn1Add(turnLeft, NULL);
     else mav_frameFn1Rmv(turnLeft, NULL);
     }
   }
 return 0;
 }

/**************************************************************************************************/

/**********************************/
/* mouse event callback for boxes */
/**********************************/
int boxMouseEvent(MAV_object *obj, MAV_mouseEvent *mouseEvent)
 {
 if (mouseEvent->movement == MAV_PRESSED) // only consider button presses
   {
   // convert from generic Maverik object to the box object
   MAV_box *b = (MAV_box *) mav_objectDataGet(obj); // can cast to a box object as the callback is defined to work only when the mouse is over a box.

   // get application specific data structure
   myStruct *ms= (myStruct *) b->userdef;

   // print contents of application specific data structure
   printf("increasing size of box number %d\n",ms->number);

   // make the box a bit bigger
   b->size.x += 1.0;
   b->size.y += 1.0;
   b->size.z += 1.0;
   }

 return 0;
 }

/***************************************/
/* mouse event callback for composites */
/***************************************/
int compositeMouseEvent(MAV_object *obj, MAV_mouseEvent *mouseEvent)
 {
 if (mouseEvent->movement == MAV_PRESSED)
   {
   MAV_composite *c = (MAV_composite *) mav_objectDataGet(obj);
   c->matrix = mav_matrixScaleSet(c->matrix, 1.1); // scale composite by a factor of 1.1
   }

 return 0;
 }

/**************************************************************************************************/

/**************************/
/* display a help message */
/**************************/
void helpMessage(void *unused)
 {
 mav_stringDisplay(mav_win_all, "Left mouse button -> navigate forward/backward and yaw", MAV_COLOUR_BLACK, 0, -0.95, 0.90);
 mav_stringDisplay(mav_win_all, "Right mouse button -> navigate up/down and left/right", MAV_COLOUR_BLACK, 0, -0.95, 0.84);
 mav_stringDisplay(mav_win_all, "Middle mouse click on cylinder -> increase radius", MAV_COLOUR_BLACK, 0, -0.95, 0.78);
 mav_stringDisplay(mav_win_all, "Middle mouse click on composite (Maverik logo) -> increase scale", MAV_COLOUR_BLACK, 0, -0.95, 0.72);
 mav_stringDisplay(mav_win_all, "e -> switch elbow prediction mode", MAV_COLOUR_BLACK, 0, -0.95, 0.66);
 mav_stringDisplay(mav_win_all, "t -> switch hand tracking", MAV_COLOUR_BLACK, 0, -0.95, 0.60);
 mav_stringDisplay(mav_win_all, "r -> reset avatar posture", MAV_COLOUR_BLACK, 0, -0.95, 0.54);
 mav_stringDisplay(mav_win_all, "  -> grip an object with mouse", MAV_COLOUR_BLACK, 0, -0.95, 0.48);
 mav_stringDisplay(mav_win_all, "w -> switch view (from head, at avatar with head movement, or at avatar no head movement)", MAV_COLOUR_BLACK, 0, -0.95, 0.42);
 mav_stringDisplay(mav_win_all, "# -> start/stop gathering training data", MAV_COLOUR_BLACK, 0, -0.95, 0.36);
 mav_stringDisplay(mav_win_all, "] -> scale avatar from file", MAV_COLOUR_BLACK, 0, -0.95, 0.30);
 mav_stringDisplay(mav_win_all, "0 -> donkey kong", MAV_COLOUR_BLACK, 0, -0.95, 0.24);
 mav_stringDisplay(mav_win_all, "q -> quit", MAV_COLOUR_BLACK, 0, -0.95, 0.18);

 mav_stringDisplay(mav_win_all, "Body control (keyboard):", MAV_COLOUR_BLACK, 0, -0.95, 0.06);
 mav_stringDisplay(mav_win_all, " cursor keys -> move head", MAV_COLOUR_BLACK, 0, -0.95, 0.00);
 mav_stringDisplay(mav_win_all, " {g b c v f d}{j n m , k l} -> {left}{right} hand controls (up down left right in out)", MAV_COLOUR_BLACK, 0, -0.95, -0.06);
 mav_stringDisplay(mav_win_all, " {s x a z}{; . / '} -> {left}{right} arm controls (up down left right)", MAV_COLOUR_BLACK, 0, -0.95, -0.12);
 mav_stringDisplay(mav_win_all, " y u i o p [ ->  torso controls (pitch roll yaw)", MAV_COLOUR_BLACK, 0, -0.95, -0.18);

 mav_stringDisplay(mav_win_all, "Movement control (mouse):", MAV_COLOUR_BLACK, 0, -0.95, -0.30);
 mav_stringDisplay(mav_win_all, " up down / left right -> walk / turn", MAV_COLOUR_BLACK, 0, -0.95, -0.36);
 mav_stringDisplay(mav_win_all, " up down / left right -> pitch / yaw head", MAV_COLOUR_BLACK, 0, -0.95, -0.42);
 }

/***************************/
/* keyboard event callback */
/***************************/
int keyEventCallback(MAV_object *obj, MAV_keyboardEvent *keyEvent)
 {
 char str1[200];
 char str2[200];
// if (keyEvent->key < 255) // only consider printable ASCII characters
//   {
 /*if (keyEvent->movement == MAV_RELEASED)
   {
   printf("key released\n");
   }*/
 if (keyEvent->movement == MAV_PRESSED)
   {
   //printf("pressed %c\n",  keyEvent->key);
   switch (keyEvent->key)
     {
     case '1':
       testingHands = !testingHands;
       if (testingHands)
         {
         printf("testing hands\n");
         initTrackers();
         mav_frameFn2Add(testHands, NULL);
         }
       else
         {
         mav_frameFn2Rmv(testHands, NULL);
         }
       break;
     case 'e': // switch between elbow position set mechanically, or by the network
       elbowPositioning = (elbowPositioning+1)%4;
       avatar->holding_left = 0;
       avatar->holding_right = 0;
       if (elbowPositioning==1)
         {
         printf("keyboard only yaw and pitch elbow position\n");
         // set arm position pointers
         leftArmRoll = &leftArmRollKeyboard;
         leftArmPitch = &leftArmPitchKeyboard;
         leftArmYaw = &leftArmYawKeyboard;
         rightArmRoll = &rightArmRollKeyboard;
         rightArmPitch = &rightArmPitchKeyboard;
         rightArmYaw = &rightArmYawKeyboard;
         // switch elbow tracking method
         mav_frameFn2Rmv(trackElbows, NULL);
         mav_frameFn2Rmv(predictElbows, NULL);
         mav_frameFn2Add(setArms, NULL);
         }
       else if (elbowPositioning==2)
         {
         printf("network predicted elbow position\n");
         // set arm position pointers
         leftArmRoll = &leftArmRollNetwork;
         leftArmPitch = &leftArmPitchNetwork;
         leftArmYaw = &leftArmYawNetwork;
         rightArmRoll = &rightArmRollNetwork;
         rightArmPitch = &rightArmPitchNetwork;
         rightArmYaw = &rightArmYawNetwork;
         // switch elbow tracking method
         mav_frameFn2Rmv(trackElbows, NULL);
         mav_frameFn2Add(predictElbows, NULL);
         mav_frameFn2Add(setArms, NULL);
         }
       else if (elbowPositioning==3)
         {
         printf("tracked elbow position\n");
         // set arm position pointers
         leftArmRoll = &leftArmRollTracked;
         leftArmPitch = &leftArmPitchTracked;
         leftArmYaw = &leftArmYawTracked;
         rightArmRoll = &rightArmRollTracked;
         rightArmPitch = &rightArmPitchTracked;
         rightArmYaw = &rightArmYawTracked;
         // switch elbow tracking method
         mav_frameFn2Rmv(predictElbows, NULL);
         mav_frameFn2Add(trackElbows, NULL);
         mav_frameFn2Add(setArms, NULL);
         }
       else
         {
         printf("mechanical elbow position\n");
         avatar->holding_left = 1;
         avatar->holding_right = 1;
         mav_frameFn2Rmv(predictElbows, NULL);
         mav_frameFn2Rmv(trackElbows, NULL);
         mav_frameFn2Rmv(setArms, NULL);
         }
       break;
     case 'w':
       viewAtHead = (viewAtHead+1)%3; // change view
       if (viewAtHead==1)
         {
         printf("view at head\n");
         mav_frameFn1Rmv(avatarView, NULL);
         mav_win_current->vp->mod = avatarModifier;
         }
       else if (viewAtHead==2)
         {
         printf("view towards avatar with HMD movement\n");
         mav_frameFn1Add(avatarView, NULL); // make view look at avatar from a distance
         mav_win_current->vp->mod = mav_TDM_vp[MAV_TDM_HMD];
         }
       else
         {
         printf("view towards avatar with no head movement\n");
         mav_frameFn1Add(avatarView, NULL); // make view look at avatar from a distance
         mav_win_current->vp->mod = NULL;
         }
       break;
     case ('t'): // make hands track mice
       if (mav_TDMModuleInit()) // try to initialise trackers
         {
         tracking = !tracking;
         if (tracking)
           {
           printf("tracking body sensors\n");
           initTrackers();
           mav_frameFn1Add(trackUpperBody, NULL);
           }
         else
           {
           printf("not tracking body sensors\n");
           mav_frameFn1Rmv(trackUpperBody, NULL);
           }
         }
       else
         printf("type tdmstart to start tracking daemon\n");
       break;
     case ('#'): // start/stop writing training data
       training = !training;
       if (training)
         {
         scanf("please enter file name of input data: %s %s\n", str1, str2);
//         handPosFileName = str1;
//         trackedElbowFileName = str2;
         if (!openTrainingFiles())
           mav_frameFn2Add(writeTrainingData, NULL); // added to phase 2 of the rendering loop (after avatars positions have been set)
         else
           {
           closeTrainingFiles();
           training = !training;
           }
         }
       else
         {
         closeTrainingFiles();
         mav_frameFn2Rmv(writeTrainingData, NULL);
         }
       break;
     case (']'):
       scanf("scaling avatar, please enter filename: %s\n", str1);
//       scanf("%s",str);
       scaleAvatar(str1);
       break;
     case ('0'):
       donkeyKong();
       break;
     case ('r'): // reset avatar position
       printf("resetting avatar position\n");
       resetPosition();
       break;
     case 'q': // exorcist quit
       eery = headYaw; // backup the head yaw (the exorcist function restores it)
       printf("aaeiiiiiiiiiiiiiiiii\n");
       mav_frameFn1Add(exorcist, NULL);
       break;
     }
   }

 switch (keyEvent->key)
   {
   case ('h'): // help
     if (keyEvent->movement == MAV_PRESSED)
       {
       // execute helpMessage function at end of frame (phase 3)
       mav_frameFn3Add(helpMessage, NULL);
       }
     else
       {
       // stop executing helpMessage function
       mav_frameFn3Rmv(helpMessage, NULL);
       }
     break;
   case (' '): // grab the object
     if (keyEvent->intersects) // check if the mouse intersects an object
       {
       if (keyEvent->movement == MAV_PRESSED) // only consider key presses
         {
         // set a pointer (objectMatrix) to the matrix field of the object.
         // this is used in the function "pick"
         if (mav_callbackGetMatrixExec(mav_win_current, keyEvent->obj, &objectMatrix))
           {
           // store distance from eye to object intersection (pt1, defined on the MAV_objectIntersection data structure objint)
           objectDistance = keyEvent->objint.pt1;
           // execute function after fixing the frames view (phase 2)
           mav_frameFn2Add(pick, NULL);
           }
         }
       else
         {
         // stop executing function
         mav_frameFn2Rmv(pick, NULL);
         }
       }
     break;
   case 'd': // left hand further away from body
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandFurther, NULL); // execute leftHandFurther function at phase 1
     else mav_frameFn1Rmv(leftHandFurther, NULL); // stop executing leftHandFurther function
     break;
   case 'f': // left hand closer in to body
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandCloser, NULL);
     else mav_frameFn1Rmv(leftHandCloser, NULL);
     break;
   case 'g': // left hand up
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandUp, NULL);
     else mav_frameFn1Rmv(leftHandUp, NULL);
     break;
   case 'b': // left hand down
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandDown, NULL);
     else mav_frameFn1Rmv(leftHandDown, NULL);
     break;
   case 'v': // left hand right
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandRight, NULL);
     else mav_frameFn1Rmv(leftHandRight, NULL);
     break;
   case 'c': // left hand left
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandLeft, NULL);
     else mav_frameFn1Rmv(leftHandLeft, NULL);
     break;
   case 'l': // right hand further away from body
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandFurther, NULL);
     else mav_frameFn1Rmv(rightHandFurther, NULL);
     break;
   case 'k': // right hand closer in to body
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandCloser, NULL);
     else mav_frameFn1Rmv(rightHandCloser, NULL);
     break;
   case 'j': // right hand up
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandUp, NULL);
     else mav_frameFn1Rmv(rightHandUp, NULL);
     break;
   case 'n': // right hand down
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandDown, NULL);
     else mav_frameFn1Rmv(rightHandDown, NULL);
     break;
   case ',': // right hand right
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandRight, NULL);
     else mav_frameFn1Rmv(rightHandRight, NULL);
     break;
   case 'm': // right hand left
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandLeft, NULL);
     else mav_frameFn1Rmv(rightHandLeft, NULL);
     break;
   case 's': // left arm pitch up
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftArmPitchUp, NULL);
     else mav_frameFn1Rmv(leftArmPitchUp, NULL);
     break;
   case 'x': // left arm pitch down
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftArmPitchDown, NULL);
     else mav_frameFn1Rmv(leftArmPitchDown, NULL);
     break;
   case 'a': // left arm yaw left
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftArmYawLeft, NULL);
     else mav_frameFn1Rmv(leftArmYawLeft, NULL);
     break;
   case 'z': // left arm yaw right
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftArmYawRight, NULL);
     else mav_frameFn1Rmv(leftArmYawRight, NULL);
     break;
   case ';': // right arm pitch up
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightArmPitchUp, NULL);
     else mav_frameFn1Rmv(rightArmPitchUp, NULL);
     break;
   case '.': // right arm pitch down
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightArmPitchDown, NULL);
     else mav_frameFn1Rmv(rightArmPitchDown, NULL);
     break;
   case '/': // left arm yaw left
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightArmYawLeft, NULL);
     else mav_frameFn1Rmv(rightArmYawLeft, NULL);
     break;
   case '\'': // left arm yaw right
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightArmYawRight, NULL);
     else mav_frameFn1Rmv(rightArmYawRight, NULL);
     break;
   case MAV_KEY_LEFT: // turn head to the left
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(headLeft, NULL);
     else mav_frameFn1Rmv(headLeft, NULL);
     break;
   case MAV_KEY_RIGHT: // turn head to the right
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(headRight, NULL);
     else mav_frameFn1Rmv(headRight, NULL);
     break;
   case MAV_KEY_UP: // tilt head up
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(headUp, NULL);
     else mav_frameFn1Rmv(headUp, NULL);
     break;
   case MAV_KEY_DOWN: // tilt head down
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(headDown, NULL);
     else mav_frameFn1Rmv(headDown, NULL);
     break;
   case 'y': // pitch torso forwards
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(torsoPitchForward, NULL);
     else mav_frameFn1Rmv(torsoPitchForward, NULL);
     break;
   case 'u': // pitch torso back
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(torsoPitchBack, NULL);
     else mav_frameFn1Rmv(torsoPitchBack, NULL);
     break;
   case 'i': // roll torso to the left
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(torsoRollLeft, NULL);
     else mav_frameFn1Rmv(torsoRollLeft, NULL);
     break;
   case 'o': // roll torso to the right
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(torsoRollRight, NULL);
     else mav_frameFn1Rmv(torsoRollRight, NULL);
     break;
   case 'p': // yaw torso right
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(torsoYawLeft, NULL);
     else mav_frameFn1Rmv(torsoYawLeft, NULL);
     break;
   case '[': // yaw torso left
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(torsoYawRight, NULL);
     else mav_frameFn1Rmv(torsoYawRight, NULL);
     break;
   }
 return 0;
 }

/**************************************************************************************************/

/******************/
/* render a frame */
/******************/
// takes a Spatial Management Structure as it's arguement (with which objects register themselves)
void renderFrame(MAV_SMS *sms1/*, MAV_SMS *sms2*/)
 {
 // check for events
 mav_eventsCheck();

 // request start of new frame
 mav_frameBegin();

  showVectors();
//  showAxis();

    drawboard();

  // display SMS (in all windows)
  mav_SMSDisplay(mav_win_all, sms1);
//  mav_SMSDisplay(mav_win_all, sms2);

 // request end of frame
 mav_frameEnd();
 }

/********/
/* main */
/********/
int main(int argc, char *argv[])
 {
// MAV_SMS *ground;
 MAV_SMS *objects;
 MAV_viewParams vp; // view parameters
 // specify TDM library
// mav_opt_TDMLib= "/opt/aig/TDM/v2.2b1/bary/lib/libtdm-polhemus.so";
 MAV_composite comp[2][6];
 int i, j;

/* Start GNU chess */

  pipe(fd);
  pipe(&fd[2]);

  if (fork()==0) {
    dup2(fd[1], 1);
    dup2(fd[2], 0);
    if (execvp("gnuchessx", NULL)) {
      fprintf(stderr, "Can't start gunchess\n");
      exit(1);
    }
  }

 // turn on multi-sampling
 mav_opt_multiSample = MAV_TRUE;

 // set stereo vision
 //mav_opt_stereo = MAV_STEREO_TWO_WINS;

 // set window sizes
 mav_opt_x = 0;
 mav_opt_y = 1152-480;
 mav_opt_width = 640;
 mav_opt_height = 480;
 mav_opt_right_x = 640;
 mav_opt_right_y = 1152-480;
 mav_opt_right_width = 640;
 mav_opt_right_height = 480;

 // initialise maverik, avatar module, and TDM module (this dynamically loads the required libraries)
 mav_initialise(&argc, argv);
 mav_avatarModuleInit();
 mav_TDMModuleInit();

 // create and populate the environment
// createEnvironment(&ground, &objects);
 /* Add a new "device" to check for events from GNU chess */

  mav_deviceNew(NULL, NULL, checkGNUchess);

/* Initialise the board */

  for (i=0; i<64; i++) board[i]= -1;
  board[sq("a1")]= (WHITE<<4) | ROOK;
  board[sq("b1")]= (WHITE<<4) | KNIGHT;
  board[sq("c1")]= (WHITE<<4) | BISHOP;
  board[sq("d1")]= (WHITE<<4) | QUEEN;
  board[sq("e1")]= (WHITE<<4) | KING;
  board[sq("f1")]= (WHITE<<4) | BISHOP;
  board[sq("g1")]= (WHITE<<4) | KNIGHT;
  board[sq("h1")]= (WHITE<<4) | ROOK;
  board[sq("a2")]= (WHITE<<4) | PAWN;
  board[sq("b2")]= (WHITE<<4) | PAWN;
  board[sq("c2")]= (WHITE<<4) | PAWN;
  board[sq("d2")]= (WHITE<<4) | PAWN;
  board[sq("e2")]= (WHITE<<4) | PAWN;
  board[sq("f2")]= (WHITE<<4) | PAWN;
  board[sq("g2")]= (WHITE<<4) | PAWN;
  board[sq("h2")]= (WHITE<<4) | PAWN;

  board[sq("a7")]= (BLACK<<4) | PAWN;
  board[sq("b7")]= (BLACK<<4) | PAWN;
  board[sq("c7")]= (BLACK<<4) | PAWN;
  board[sq("d7")]= (BLACK<<4) | PAWN;
  board[sq("e7")]= (BLACK<<4) | PAWN;
  board[sq("f7")]= (BLACK<<4) | PAWN;
  board[sq("g7")]= (BLACK<<4) | PAWN;
  board[sq("h7")]= (BLACK<<4) | PAWN;
  board[sq("a8")]= (BLACK<<4) | ROOK;
  board[sq("b8")]= (BLACK<<4) | KNIGHT;
  board[sq("c8")]= (BLACK<<4) | BISHOP;
  board[sq("d8")]= (BLACK<<4) | QUEEN;
  board[sq("e8")]= (BLACK<<4) | KING;
  board[sq("f8")]= (BLACK<<4) | BISHOP;
  board[sq("g8")]= (BLACK<<4) | KNIGHT;
  board[sq("h8")]= (BLACK<<4) | ROOK;

/* Read in AC3D description of pieces  */

  mav_compositeReadAC3D("chess pieces/pawn.wh.ac", &comp[WHITE][PAWN], MAV_ID_MATRIX);
  mav_compositeReadAC3D("chess pieces/rook.wh.ac", &comp[WHITE][ROOK], MAV_ID_MATRIX);
  mav_compositeReadAC3D("chess pieces/knight.wh.ac", &comp[WHITE][KNIGHT], MAV_ID_MATRIX);
  mav_compositeReadAC3D("chess pieces/bishop.wh.ac", &comp[WHITE][BISHOP], MAV_ID_MATRIX);
  mav_compositeReadAC3D("chess pieces/queen.wh.ac", &comp[WHITE][QUEEN], MAV_ID_MATRIX);
  mav_compositeReadAC3D("chess pieces/king.wh.ac", &comp[WHITE][KING], MAV_ID_MATRIX);

  mav_compositeReadAC3D("chess pieces/pawn.bl.ac", &comp[BLACK][PAWN], MAV_ID_MATRIX);
  mav_compositeReadAC3D("chess pieces/rook.bl.ac", &comp[BLACK][ROOK], MAV_ID_MATRIX);
  mav_compositeReadAC3D("chess pieces/knight.bl.ac", &comp[BLACK][KNIGHT], MAV_ID_MATRIX);
  mav_compositeReadAC3D("chess pieces/bishop.bl.ac", &comp[BLACK][BISHOP], MAV_ID_MATRIX);
  mav_compositeReadAC3D("chess pieces/queen.bl.ac", &comp[BLACK][QUEEN], MAV_ID_MATRIX);
  mav_compositeReadAC3D("chess pieces/king.bl.ac", &comp[BLACK][KING], MAV_ID_MATRIX);

/* Register them as Maverik objects */

  for (i=0; i<2; i++) {
    for (j=0; j<6; j++) {
      piece[i][j]= mav_objectNew(mav_class_composite, &comp[i][j]);
    }
  }

/* Set perspective, background colour and backface culling */

  mav_windowPerspectiveSet(mav_win_all, 0.1, 1000, 70.0, 1.25);
  mav_windowBackgroundColourSet(mav_win_all, 0.0, 0.5, 1.0);
  mav_windowBackfaceCullSet(mav_win_all, MAV_TRUE);
printf("test\n");

 createAvatar(&objects);
printf("test2\n");

 mav_stp_default.offset = 0.065; // set the stereo offset (must be done after initialising maverik)

 // set up viewing parameters (needed to prevent run-time warning)
 vp.eye = mav_matrixXYZGet(avatar->matrix);
 vp.view = mav_vectorSet(0,0,-1);
 vp.right = MAV_X_VECTOR;
 vp.up = MAV_Y_VECTOR;
 vp.mod = avatarModifier; // modify eye view using avatar head position

 // use these view parameters in all windows
 mav_windowViewParamsSet(mav_win_all, &vp);

 // default mouse navigation
 mav_navigationMouse(mav_win_all, mav_navigationMouseDefault);
 // customize behaviour
 mav_navigationMouseDefaultParams(mav_win_all, MAV_LEFT_BUTTON, turn, 0.01, -0.0005, walk, 0.01, 0.0005);
 mav_navigationMouseDefaultParams(mav_win_all, MAV_RIGHT_BUTTON, yawHead, 0.01, -0.0005, pitchHead, 0.01, -0.0005);

 // also use default keyboard navigation
 //mav_navigationKeyboard(mav_win_all, mav_navigationKeyboardDefault);

 // define mouse callback (the same for all windows, when the mouse is over a box).
 // could also use mav_class_any for when the mouse over any object, or mav_class_miss for when the mouse is over no object.
 mav_callbackMouseSet(MAV_MIDDLE_BUTTON, mav_win_all, mav_class_box, boxMouseEvent);
 // define mouse callback for composite objects
 mav_callbackMouseSet(MAV_MIDDLE_BUTTON, mav_win_all, mav_class_composite, compositeMouseEvent);

 // define keyboard event callback (the same for all windows, when the mouse is anywhere in the world)
 mav_callbackKeyboardSet(mav_win_all, mav_class_world, keyEventCallback);

 // define TDM callback
 mav_callbackTDMSet(mav_win_all, mav_class_world, tdmEventCallback);

 // start rendering loop
 while(1) renderFrame(objects);

 }

/**************************************************************************************************/

