// created: 2.4.01 modified: 23.8.01 final clean-up: 24.4.02
// Stephan Dale
// MAVERIK (v6.2) avatar movement program

//#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "maverik.h"
#include "mav_avatar.h"
#include "mavlib_avatar.h" // for avatar parts definitions
#include "mav_tdm.h" // for TDM callback

#include "sjd_backprop.h" // for the NN datatype
#include "avatarlib.h"
#include "avatarlibio.h"

#define NORMAL 1
#define TRAINING 2
#define TESTING 3

/**************************************************************************************************/

/**********************/
/* external variables */
/**********************/

// from avatarlib
extern MAV_avatar *avatar; // pointer to the avatar

extern int holdingLeft;
extern int holdingRight;
extern MAV_matrix leftHandMatrix; // left hand world position
extern MAV_matrix rightHandMatrix; // ditto right hand

// avatar positional data
extern float avatarYaw, headRoll, headPitch, headYaw, torsoRoll, torsoPitch, torsoYaw;
extern float *leftArmRoll, *leftArmPitch, *leftArmYaw, *rightArmRoll, *rightArmPitch, *rightArmYaw;
extern MAV_vector *eyePoint; // eye point relative to back

// (set) avatar data
extern MAV_vector eyePointAvatar;
// keyboard data
extern float leftArmRollKeyboard, leftArmPitchKeyboard, leftArmYawKeyboard, rightArmRollKeyboard, rightArmPitchKeyboard, rightArmYawKeyboard;
// tracked data
extern MAV_vector eyePointTracked;
extern float leftArmRollTracked, leftArmPitchTracked, leftArmYawTracked, rightArmRollTracked, rightArmPitchTracked, rightArmYawTracked;
// neural network data
extern float leftArmRollNetwork, leftArmPitchNetwork, leftArmYawNetwork, rightArmRollNetwork, rightArmPitchNetwork, rightArmYawNetwork;

// training and testing data file names
extern char trainingFilename[200], testingFilename[200];

/**************************************************************************************************/

/********************/
/* global variables */
/********************/

// environment variables
MAV_SMS *ground;
MAV_SMS *objects;
MAV_rectangle rect; // ground plane
MAV_composite comp;
MAV_sphere sphere[343]; // 7x7x7 for random arm movements
MAV_cylinder cyl[5]; // for rotation arm movements
MAV_sphere hole; // ditto

int programMode = NORMAL; // indicates whether using normally, training, or testing
int view = 1; // specify whether view at avatars head, or looking at avatar
int elbowMode = 1; // specify keyboard yaw and pitch, neural network prediction, tracked, or old MAVERIK mechanical method
int tracking = 0; // whether hands are to track the 3D mice
int writing = 0; // indicates whether writing data to files or not
//int keyboardControl = 1; // switch keyboard control of avatar
int vectorDisplay = 0; // switch between nothing, arm vectors (predicted and corrected), arm vectors and NN predicted elbow, arm vectors and tracked elbow.
int netLoaded = 0; // indicates if a NN is loaded
int environment = 0; // specifies environment loaded
int eery = 0; // exorcist variable

int testingHands = 0;
MAV_matrix *rightObjMat, *leftObjMat; // pointer to the matrix of the object being held (if there is one)

/***********************/
/* function prototypes */
/***********************/

void printUsage(void);

void calcPosRPY(MAV_matrix mat, float *roll, float *pitch, float *yaw);

void defineRectangle(MAV_rectangle *r);
void defineComposite(MAV_composite *c);
void setSphere(MAV_sphere *sphere, float radius, MAV_vector position);
void setCylinder(MAV_cylinder *cyl, MAV_vector position);

//void deleteOnIntersect(void *unused);
int pick(int hand);
void drop(int hand);
void blackHole(void *unused);
void holdRight(void *unused);
void holdLeft(void *unused);

void exorcist(void *unused);

int sphereMouseEvent(MAV_object *obj, MAV_mouseEvent *mouseEvent);
int compositeMouseEvent(MAV_object *obj, MAV_mouseEvent *mouseEvent);
int tdmEventCallback(MAV_object *obj, MAV_TDMEvent *tdmEvent);

void help(void);
void helpMessage(void *unused);
void switchTracking(void);
void switchView(void);
void switchWriting(void);
int keyEventCallback(MAV_object *obj, MAV_keyboardEvent *keyEvent);

void createMainEnvironment(MAV_SMS **groundPlane, MAV_SMS **theObjects);
void createEnvironment1(MAV_SMS **theObjects);
void deleteEnvironment1(MAV_SMS **theObjects);
void createEnvironment2(MAV_SMS **groundPlane, MAV_SMS **theObjects);
void deleteEnvironment2(MAV_SMS **groundPlane, MAV_SMS **theObjects);
void createEnvironment3(MAV_SMS **groundPlane, MAV_SMS **theObjects);
void deleteEnvironment3(MAV_SMS **groundPlane, MAV_SMS **theObjects);

void showLines(void);

void renderFrame(MAV_SMS *sms1, MAV_SMS *sms2);
int main(int argc, char *argv[]);

/**************************************************************************************************/

/**************/
/* show usage */
/**************/
void printUsage(void)
 {
 printf("\nUsage: main [option] [filenames]\n\n");

 printf("[option] -n normal use\n");
 printf("         -t gather training data\n");
 printf("         -e gather testing data\n\n");

 printf("[filenames] with -n: <user> <neural network>\n");
 printf("            with -t: <user> <training data>\n");
 printf("            with -e: <user> <neural network> <testing data>\n\n");

 printf("<user> location of user dimensions file\n");
 printf("<neural network> location of neural network (NN) file\n");
 printf("<training data> location to write training data\n");
 printf("<testing data> location to write testing data\n\n");

 printf("Additional info: Only one switch can be used at a time.\n");
 printf("                 If no switches are given the program defaults to normal use mode.\n");
 printf("                 Different NN, user, and data filenames can be specified when the program is running.\n");
 printf("                 Writing of data must be started/stopped when the program is running.\n");
 printf("                 To see a list of commands, start the program 'main' and press 'h' for help.\n\n");

 printf("Examples: Normal use: 'main -n users/caz nets/network1'\n");
 printf("          Training: 'main -t users/caz data/traindata1'\n");
 printf("          Testing: 'main -e users/caz nets/network1 data/testdata1'\n\n");
 }

/**************************************************************************************************/

/*******************************************************/
/* function to calculate positive roll, pitch, and yaw */
/*******************************************************/
void calcPosRPY(MAV_matrix mat, float *roll, float *pitch, float *yaw)
 {
 MAV_vector vec, vec2;
 float r, p, y;
 // get roll, pitch, and yaw
 mav_matrixRPYGet(mat, &r, &p, &y);

 // calculate direction vector
 vec = mav_vectorMult(MAV_X_VECTOR, mav_matrixSet(r, p, y, 0, 0, 0));
 // calculate positive roll..
 vec2 = vec;
 vec2.z = 0; // put vec onto xy plane.
 if (mav_vectorMag(vec2)!=0)
   {
   *roll = MAV_RAD2DEG(acos(mav_vectorDotProduct(MAV_X_VECTOR,vec2)/mav_vectorMag(vec2)));
   if (vec2.x < 0){ *roll = -(*roll); }
   }
 else *roll = 0;

 // ..yaw..
 vec2 = vec;
 vec2.y = 0; // put vec onto xz plane
 if (mav_vectorMag(vec2)!=0)
   {
   *yaw = MAV_RAD2DEG(acos(mav_vectorDotProduct(MAV_X_VECTOR,vec2)/mav_vectorMag(vec2)));
   if (vec2.z < 0){ *yaw = -(*yaw); }
   }
 else *yaw = 0;

 // ..and pitch
 vec2 = vec;
 vec2.x = 0; // put vec onto yz plane
 if (mav_vectorMag(vec2)!=0)
   {
   *pitch = MAV_RAD2DEG(acos(mav_vectorDotProduct(MAV_Z_VECTOR,vec2)/mav_vectorMag(vec2)));
   if (vec2.y < 0){ *pitch = -(*pitch); }
   }
 else *pitch = 0;

 printf("roll=%f pitch=%f yaw=%f\n",*roll,*pitch,*yaw);

 }

/**************************************************************************************************/

/**********************/
/* define a rectangle */
/**********************/
void defineRectangle(MAV_rectangle *r)
 {
 // size
 r->width = 50.0;
 r->height = 50.0;

 // texture tiling
 r->xtile = 1;
 r->ytile = 1;

 r->sp = mav_surfaceParamsNew(MAV_TEXTURE, 0, 0, 5); // decal texture index 5 (defined in main).

 // orientation (roll,pitch,yaw) and position (x,y,z).
 r->matrix = mav_matrixSet(0, -90, 0, 0, 0, 0); // the avatar is 1.7 tall (or should I say -1.7 short)
 }

/*****************************/
/* define a composite object */
/*****************************/
void defineComposite(MAV_composite *c)
 {
 // read AC3D object from file
 if (!mav_compositeReadAC3D("mavlogo.ac", c, MAV_ID_MATRIX))
   {
   printf("failed to read mavlogo.ac\n");
   exit(1);
   }

 c->matrix= mav_matrixSet(0, 0, 0, 0, 1.7, -10);
 }

/***********************************/
/* define a sphere and position it */
/***********************************/
void setSphere(MAV_sphere *sphere, float radius, MAV_vector position)
 {
 // size
 sphere->radius = radius;

 // surface parameters.
 sphere->sp = mav_surfaceParamsNew(MAV_MATERIAL,0,4,0); // material index 4

 // position sphere
 sphere->matrix = mav_matrixXYZSet(MAV_ID_MATRIX, position);

}

/*************************************/
/* define a cylinder and position it */
/*************************************/
void setCylinder(MAV_cylinder *c, MAV_vector position)
 {
 // size
 c->radius = 0.025;
 c->height = 0.1;
 c->endcap = 1;
 c->nverts = 10;

 c->sp = mav_surfaceParamsNew(MAV_LIT_TEXTURE,  0, 4, 0);

 // position
 c->matrix = mav_matrixXYZSet(MAV_ID_MATRIX, position);
 c->matrix = mav_matrixRPYSet(c->matrix, 0,90,0);
 }

/**************************************************************************************************/

/**************************************************/
/* check right hand for intersection with spheres */
/**************************************************/
void deleteOnIntersect(void *unused)
 {
 MAV_SMS *tmpSMS;
 MAV_BB BB, handBB;

 mav_callbackBBExec(mav_win_all, avatar->part[RIGHT_HAND]->obj, &BB); // get hand bounding box

 // align bounding box
 mav_BBAlign(BB, rightHandMatrix, &handBB);

 tmpSMS = mav_SMSObjListNew();
 if (mav_SMSIntersectBB(mav_win_current, objects, handBB, tmpSMS)) // if any of the spheres intersect the hand
   {
   mav_SMSDelete(tmpSMS, MAV_TRUE); // delete the sms and it's contents
   }
 }

/**************/
/* black hole */
/**************/
void blackHole(void *unused)
 {
 MAV_SMS *tmpSMS;
 MAV_BB BB, holeBB;

 mav_sphereBB(mav_objectNew(mav_class_sphere, &hole), &holeBB); // get black hole bounding box

 tmpSMS = mav_SMSObjListNew();
 if (mav_SMSIntersectBB(mav_win_current, objects, holeBB, tmpSMS)) // if any of the objects intersect the black hole
   {
   mav_SMSDelete(tmpSMS, MAV_TRUE); // delete the sms and it's contents
   }

 }

/*************************************/
/* pick up an object with right hand */
/*************************************/
int pick(int hand)
 {
 MAV_SMS *tmpSMS;
 MAV_object *obj;
 MAV_BB BB, handBB;

 if (hand == RIGHT_HAND)
   if (rightObjMat != NULL) return 0; // if already holding cannot pick up other objects, so return fail
 else
   if (leftObjMat != NULL) return 0;

 mav_callbackBBExec(mav_win_all, avatar->part[hand]->obj, &BB); // get hand bounding box

  // align bounding box
 if (hand==RIGHT_HAND)
   mav_BBAlign(BB, rightHandMatrix, &handBB);
 else
   mav_BBAlign(BB, leftHandMatrix, &handBB);

 tmpSMS = mav_SMSObjListNew();
 if (mav_SMSIntersectBB(mav_win_current, objects, handBB, tmpSMS)) // if any of the spheres intersect the hand
   {
   while (mav_SMSCallbackObjectNextExec(tmpSMS, &obj)) // go through objects in SMS
     {
     // set the objects matrix to the matrix of the hand
     if (hand == RIGHT_HAND)
       {
       mav_callbackGetMatrixExec(mav_win_current, obj, &rightObjMat); // get objects matrix
       mav_frameFn2Add(holdRight, NULL); // hold the object
       }
     else
       {
       mav_callbackGetMatrixExec(mav_win_current, obj, &leftObjMat); // get objects matrix
       mav_frameFn2Add(holdLeft, NULL);
       }
     }
   }

 return 1; // success
 }

void drop(int hand)
 {
 if (hand == RIGHT_HAND)
   {
   if (rightObjMat != NULL);
     {
     mav_frameFn2Rmv(holdRight, NULL);
     rightObjMat = NULL;
     }
   }
 else
   {
   if (leftObjMat != NULL)
     {
     mav_frameFn2Rmv(holdLeft, NULL);
     leftObjMat = NULL;
     }
   }
 }

void holdRight(void *unused)
 {
 *rightObjMat = mav_matrixMult(rightHandMatrix, mav_matrixXYZSet(MAV_ID_MATRIX, mav_vectorSet(-0.025,-0.08,0)));  // shift to middle of hand from wrist
 }

void holdLeft(void *unused)
 {
 *leftObjMat = mav_matrixMult(leftHandMatrix, mav_matrixXYZSet(MAV_ID_MATRIX, mav_vectorSet(-0.025,-0.08,0)));  // shift to middle of hand from wrist
 }

/**************************************************************************************************/

/************/
/* exorcist */
/************/
void exorcist(void *unused)
 {
 if ((tracking!=1) & (headYaw < (eery+359))) // rotate head (and keep vertical)
   {
   headYaw += 10;
   }
 else // stop rotating head, and reset previous head position
   {
   headYaw = eery;
   mav_frameFn1Rmv(exorcist, NULL);
   exit(1);
   }
 }

/**************************************************************************************************/

/**********************/
/* TDM event callback */
/**********************/

int tdmEventCallback(MAV_object *obj, MAV_TDMEvent *tdmEvent)
 {
 MAV_BB handBB;

 if (tdmEvent->tracker == MAV_TDM_BLUE)
   {
   // pick/drop object
   if (tdmEvent->button == MAV_TDM_MIDDLE) 
     {
     if (tdmEvent->movement == MAV_PRESSED) pick(RIGHT_HAND); //mav_frameFn2Add(pickRight, NULL); // hold object
     else drop(RIGHT_HAND); //mav_frameFn2Rmv(pickRight, NULL); // release object
     }
   // move forwards if top button of blue hand sensor is pressed
   else if (tdmEvent->button == MAV_TDM_TOP)
     {
     if (programMode!=NORMAL)
       {
       if (tdmEvent->movement == MAV_PRESSED) switchWriting(); // start/stop writing data
       }
     else
       {
       if (tdmEvent->movement == MAV_PRESSED) mav_frameFn1Add(moveForwards, NULL);
       else mav_frameFn1Rmv(moveForwards, NULL);
       }
     }
   // move backwards if bottom button of blue hand sensor is pressed
   else if (tdmEvent->button == MAV_TDM_BOTTOM) 
     {
     if (programMode!=NORMAL)
       {
       if (tdmEvent->movement == MAV_PRESSED) switchTracking(); // turn tracking on/off
       }
     else
       {
       if (tdmEvent->movement == MAV_PRESSED) mav_frameFn1Add(moveBackwards, NULL);
       else mav_frameFn1Rmv(moveBackwards, NULL);
       }
     }
   }
 else if (tdmEvent->tracker == MAV_TDM_RED)
   {
   // initialise sensors if middle button of red hand sensor is pressed
   if (tdmEvent->button == MAV_TDM_MIDDLE)
     {
     if (tdmEvent->movement == MAV_PRESSED) pick(LEFT_HAND); //mav_frameFn2Add(pickLeft, NULL); // hold object
     else drop(LEFT_HAND); //mav_frameFn2Rmv(pickLeft, NULL); // release object
     }
   // turn right if top button of red hand sensor is pressed
   else if (tdmEvent->button == MAV_TDM_TOP)
     {
     if (programMode!=NORMAL)
       {
       if (tdmEvent->movement == MAV_PRESSED) switchTracking(); // turn tracking on/off
       }
     else
       {
       if (tdmEvent->movement == MAV_PRESSED) mav_frameFn1Add(turnRight, NULL);
       else mav_frameFn1Rmv(turnRight, NULL);
       }
     }
   // turn left if bottom button of red hand sensor is pressed
   else if (tdmEvent->button == MAV_TDM_BOTTOM) 
     {
     if (programMode!=NORMAL)
       {
       if (tdmEvent->movement == MAV_PRESSED) switchView(); // change view
       }
     else
       {
       if (tdmEvent->movement == MAV_PRESSED) mav_frameFn1Add(turnLeft, NULL);
       else mav_frameFn1Rmv(turnLeft, NULL);
       }
     }
   }
 return 0;
 }

/**************************************************************************************************/

/**********************************/
/* mouse event callback for boxes */
/**********************************/
/*int boxMouseEvent(MAV_object *obj, MAV_mouseEvent *mouseEvent)
 {
 if (mouseEvent->movement == MAV_PRESSED) // only consider button presses
   { 
   // convert from generic Maverik object to the box object
   MAV_box *b = (MAV_box *) mav_objectDataGet(obj); // can cast to a box object as the callback is defined to work only when the mouse is over a box.

   // get application specific data structure
   myStruct *ms= (myStruct *) b->userdef;

   // print contents of application specific data structure
   printf("increasing size of box number %d\n",ms->number);

   // make the box a bit bigger
   b->size.x += 1.0;
   b->size.y += 1.0;
   b->size.z += 1.0;
   }     

 return 0;
 }*/

/***************************************/
/* mouse event callback for composites */
/***************************************/
int compositeMouseEvent(MAV_object *obj, MAV_mouseEvent *mouseEvent)
 {
 if (mouseEvent->movement == MAV_PRESSED) 
   {
   MAV_composite *c = (MAV_composite *) mav_objectDataGet(obj);
   c->matrix = mav_matrixScaleSet(c->matrix, 1.1); // scale composite by a factor of 1.1
   }

 return 0;
 }

/**************************************************************************************************/

/**************************/
/* display a help message */
/**************************/
void help(void) // for the shell
 {
 printf("\n--- help ---\n\n");

 printf("ALL MODES:\n");
 printf(" - -> switch program mode (normal use, training, or testing)\n");
 printf(" ] -> scale avatar from file\n");
 printf(" # -> load network\n");
 printf(" } -> set sensor offsets\n");
 printf(" t -> switch tracking\n");
 printf(" w -> switch view\n");
 printf(" e -> switch elbow display (keyboard, NN, tracked, or MAVERIK default)\n");
 printf(" | -> switch arm vectors (arm, arm & NN elbow, arm & tracked elbow)\n");
 printf(" 8 -> donkey kong\n");
 printf(" h -> help\n");
 printf("   -> grip an object with mouse\n");
 printf(" q -> quit\n\n");

 printf("NORMAL MODE ONLY:\n");
 printf(" r -> reset avatar posture\n\n");

 printf("TRAINING AND TESTING MODES ONLY:\n");
 printf(" @ -> set file names\n");
 printf(" = -> start/stop writing data\n\n");

 printf("BODY CONTROL (hand and arm controls disabled in training or testing mode):\n");
 printf(" cursor keys -> move head\n");
 printf(" {g b c v f d}{j n m , k l} -> {left}{right} arm controls (up down left right in out)\n");
 printf(" {G B C V D F}{J N M < K L} -> {left}{right} hand controls (pitch-up pitch-down yaw-left yaw-right roll-left roll-right)\n");
 printf(" {s x a z}{; . / '} -> {left}{right} elbow controls (pitch-up pitch-down yaw-left yaw-right)\n");
 printf(" y u i o p [ ->  torso controls (pitch roll yaw)\n\n");

 printf("MOVEMENT CONTROL (mouse):\n");
 printf(" with left button held, up down / left right -> walk / turn\n");
 printf(" with right button held, up down / left right -> pitch / yaw head\n\n");

 printf("ENVIRONMENT:\n");
 printf(" {0 1 2 3} -> change environment\n");
 printf(" Middle mouse click on cylinder -> increase radius\n");
 printf(" Middle mouse click on composite (Maverik logo) -> increase scale\n\n");

 printf("3D MICE CONTROL:\n");
 printf(" thumb buttons -> grab object\n");
 printf(" in normal mode, triggers BLUE / RED -> walk / turn\n");
 printf(" in training or testing mode, BLUE top trigger -> switch writing\n");
 printf("                              BLUE bottom trigger -> switch tracking\n");
 
 printf("\n--- help ---\n\n");
 }

void helpMessage(void *unused) // for the VR window
 {
 mav_stringDisplay(mav_win_all, "ALL MODES:", MAV_COLOUR_BLACK, 0, -0.95, 0.95);
 mav_stringDisplay(mav_win_all, " - -> switch program mode (normal use, training, or testing)", MAV_COLOUR_BLACK, 0, -0.95, 0.90);
 mav_stringDisplay(mav_win_all, " ] -> scale avatar from file", MAV_COLOUR_BLACK, 0, -0.95, 0.85);
 mav_stringDisplay(mav_win_all, " # -> load network", MAV_COLOUR_BLACK, 0, -0.95, 0.80);
 mav_stringDisplay(mav_win_all, " } -> set sensor offsets", MAV_COLOUR_BLACK, 0, -0.95, 0.75);
 mav_stringDisplay(mav_win_all, " t -> switch tracking", MAV_COLOUR_BLACK, 0, -0.95, 0.70);
 mav_stringDisplay(mav_win_all, " w -> switch view", MAV_COLOUR_BLACK, 0, -0.95, 0.65);
 mav_stringDisplay(mav_win_all, " e -> switch elbow display (keyboard, NN, tracked, or MAVERIK default)", MAV_COLOUR_BLACK, 0, -0.95, 0.60);
 mav_stringDisplay(mav_win_all, " | -> switch arm vectors (arm, arm & NN elbow, arm & tracked elbow)", MAV_COLOUR_BLACK, 0, -0.95, 0.55);
 mav_stringDisplay(mav_win_all, " 8 -> donkey kong", MAV_COLOUR_BLACK, 0, -0.95, 0.50);
 mav_stringDisplay(mav_win_all, " h -> help", MAV_COLOUR_BLACK, 0, -0.95, 0.45);
 mav_stringDisplay(mav_win_all, "   -> grip an object (with avatar's hands)", MAV_COLOUR_BLACK, 0, -0.95, 0.40);
 mav_stringDisplay(mav_win_all, " q -> quit", MAV_COLOUR_BLACK, 0, -0.95, 0.35);

 mav_stringDisplay(mav_win_all, "NORMAL MODE ONLY:", MAV_COLOUR_BLACK, 0, -0.95, 0.28);
 mav_stringDisplay(mav_win_all, " r -> reset avatar posture", MAV_COLOUR_BLACK, 0, -0.95, 0.23);

 mav_stringDisplay(mav_win_all, "TRAINING AND TESTING MODES ONLY:", MAV_COLOUR_BLACK, 0, -0.95, 0.16);
 mav_stringDisplay(mav_win_all, " @ -> set file names", MAV_COLOUR_BLACK, 0, -0.95, 0.11);
 mav_stringDisplay(mav_win_all, " = -> start/stop writing data", MAV_COLOUR_BLACK, 0, -0.95, 0.06);

 mav_stringDisplay(mav_win_all, "BODY CONTROL (hand and arm controls disabled in training or testing mode):", MAV_COLOUR_BLACK, 0, -0.95, -0.01);
 mav_stringDisplay(mav_win_all, " cursor keys -> move head", MAV_COLOUR_BLACK, 0, -0.95, -0.06);
 mav_stringDisplay(mav_win_all, " {g b c v f d}{j n m , k l} -> {left}{right} arm controls (up down left right in out)", MAV_COLOUR_BLACK, 0, -0.95, -0.11);
 mav_stringDisplay(mav_win_all, " {G B C V D F}{J N M < K L} -> {left}{right} hand controls (pitch-up pitch-down yaw-left yaw-right roll-left roll-right)", MAV_COLOUR_BLACK, 0, -0.95, -0.16);
 mav_stringDisplay(mav_win_all, " {s x a z}{; . / '} -> {left}{right} elbow controls (pitch-up pitch-down yaw-left yaw-right)", MAV_COLOUR_BLACK, 0, -0.95, -0.21);
 mav_stringDisplay(mav_win_all, " y u i o p [ ->  torso controls (pitch roll yaw)", MAV_COLOUR_BLACK, 0, -0.95, -0.26);

 mav_stringDisplay(mav_win_all, "MOVEMENT CONTROL (mouse):", MAV_COLOUR_BLACK, 0, -0.95, -0.33);
 mav_stringDisplay(mav_win_all, " with left button held, up down / left right -> walk / turn", MAV_COLOUR_BLACK, 0, -0.95, -0.38);
 mav_stringDisplay(mav_win_all, " with right button held, up down / left right -> pitch / yaw head", MAV_COLOUR_BLACK, 0, -0.95, -0.43);

 mav_stringDisplay(mav_win_all, "ENVIRONMENT:", MAV_COLOUR_BLACK, 0, -0.95, -0.50);
 mav_stringDisplay(mav_win_all, " {0 1 2 3} -> change environment", MAV_COLOUR_BLACK, 0, -0.95, -0.55);
 mav_stringDisplay(mav_win_all, " Middle mouse click on cylinder -> increase radius", MAV_COLOUR_BLACK, 0, -0.95, -0.60);
 mav_stringDisplay(mav_win_all, " Middle mouse click on composite (Maverik logo) -> increase scale", MAV_COLOUR_BLACK, 0, -0.95, -0.65);

 mav_stringDisplay(mav_win_all, "3D MICE CONTROL:", MAV_COLOUR_BLACK, 0, -0.95, -0.72);
 mav_stringDisplay(mav_win_all, " thumb buttons -> grab object", MAV_COLOUR_BLACK, 0, -0.95, -0.77);
 mav_stringDisplay(mav_win_all, " in normal mode, triggers BLUE / RED -> walk / turn", MAV_COLOUR_BLACK, 0, -0.95, -0.82);
 mav_stringDisplay(mav_win_all, " in training or testing mode, BLUE top trigger -> switch writing", MAV_COLOUR_BLACK, 0, -0.95, -0.87);
 mav_stringDisplay(mav_win_all, " in training or testing mode, BLUE bottom trigger -> switch tracking", MAV_COLOUR_BLACK, 0, -0.95, -0.92);
 
 }

/************************/
/* turn tracking on/off */
/************************/

void switchTracking(void)
 {
 if (mav_TDMModuleInit())
   {
   // define TDM callback
   mav_callbackTDMSet(mav_win_all, mav_class_world, tdmEventCallback);
   /*if (programMode==NORMAL)*/ tracking = !tracking; // only allow normal mode to switch sensors off
   /*else if (!tracking) tracking==1; // training and testing mode only attempt to switch tracking on (if it's currently off)*/
   if (tracking)
     { 
     printf("tracking on\n");
     initTrackers(FRONT);
     mav_frameFn1Add(trackUpperBody, NULL);
     if (view==1 || view==3) eyePoint = &eyePointTracked; // only track eye point if using a view that moves with head
     else eyePoint = &eyePointAvatar;
     }
   else
     {
     printf("tracking off\n");
     mav_frameFn1Rmv(trackUpperBody, NULL);
     eyePoint = &eyePointAvatar; // make sure that the tracked eye point is not used
     }
   }
 else
   printf("cannot track body:- type 'tdmstart' in relevant directory to start tracking daemon\n");
 }

/************************/
/* switch between views */
/************************/
void switchView()
 {
 view = (view+1)%5;
 // set eye point
 if (tracking) eyePoint = &eyePointTracked;
 else eyePoint = &eyePointAvatar;
 if (view==1)
   {
   printf("view at head\n");
   mav_frameFn1Rmv(fixHeadLookingForward, NULL); // allow head movement
   mav_win_current->vp->mod = avatarView;
   }
 else if (view==2)
   {
   printf("view at head with torso movement (but no head movement)\n");
   eyePoint = &eyePointAvatar; // ensure eye point is not tracked
   mav_frameFn1Add(fixHeadLookingDown, NULL); // fix head looking down
   mav_win_current->vp->mod = avatarView;
   }
 else if (view==3)
   {
   printf("view floating far from avatar with HMD movement\n");
   mav_frameFn1Rmv(fixHeadLookingDown, NULL); // allow head movement
   mav_win_current->vp->mod = mav_TDM_vp[MAV_TDM_HMD];
   }
 else if (view==4)
   {
   printf("fixed view floating far from avatar\n");
   mav_win_current->vp->mod = floatingFarView;
   }
 else
   {
   printf("fixed view floating near avatar (and no head movement)\n");
   mav_frameFn1Add(fixHeadLookingForward, NULL); // prevent the head from moving
   mav_win_current->vp->mod = floatingNearView;
   }
 }

/**********************/
/* start writing data */
/**********************/
void switchWriting(void)
 {
 writing = !writing;
 if (programMode==NORMAL)
   {
   writing = 0;
   printf("cannot write in normal mode:- press '-' to switch to training or testing mode\n");
   }
 else if (programMode==TRAINING) // if in training mode...
   {
   if (writing) //...and writing, try to open files and start writing to them
     {
     if (openTrainingFiles()) // if you can open files, start writing
       mav_frameFn3Add(writeTrainingData, NULL); // added to phase 2 of the rendering loop (after avatars positions have been set)
     else writing = !writing; // else reset
     }
   else // if not writing any more, stop writing data
     {
     mav_frameFn3Rmv(writeTrainingData, NULL);
     closeTrainingFile();
     }
   }
 else if (programMode==TESTING) // if testing mode
   {
   if (writing) //...and writing, try to open files and start writing to them
     {
     if (openTestingFiles()) // if you can open files, start writing
       mav_frameFn3Add(writeTestingData, NULL); // added to phase 2 of the rendering loop (after avatars positions have been set)
     else writing = !writing; // else reset
     }
   else // if not writing any more, stop writing data
     {
     mav_frameFn3Rmv(writeTestingData, NULL);
     closeTestingFiles();
     }
   }
 else
   {
   writing = 0;
   printf("cannot write:- press '-' to switch to training or testing mode\n");
   }

 }

/***************************/
/* keyboard event callback */
/***************************/
int keyEventCallback(MAV_object *obj, MAV_keyboardEvent *keyEvent)
 {
 char str1[200];
 char str2[200];

// if (keyEvent->key < 255) // only consider printable ASCII characters
//   {

 // key switch for press-key actions
 if (keyEvent->movement == MAV_PRESSED)
   {
   //printf("pressed %c\n",  keyEvent->key);
   switch (keyEvent->key)
     {
     case '~':
       testingHands = !testingHands;
       if (testingHands)
         {
         printf("testing hands\n");
         initTrackers(FRONT);
         mav_frameFn2Add(testHands, NULL);
         }
       else
         {
         mav_frameFn2Rmv(testHands, NULL);
         }
       break;
     case '-': // change program mode
       programMode = (programMode+1)%3;
       if (programMode==NORMAL) // normal use
         {
         if (!netLoaded) printf("warning:- a network is not loaded, press '#' to load one\n");
         if (writing) // if writing, stop
           {
           mav_frameFn3Rmv(writeTestingData, NULL); // stop any testing
           writing = 0; // make sure writing stops
           closeTestingFiles();
           }
         printf("normal use mode\n");
         holdingLeft = 1;
         mav_frameFn2Rmv(trackElbow, NULL);
         resetPosition();
         mav_frameFn2Add(predictElbows, NULL);
         }
       else if (programMode==TRAINING) // training mode
         {
         if (!tracking) printf("warning:- not tracking, press 't' to switch tracking on\n");
         // previous (normal) mode does not write so don't need to test for writing
         printf("training mode\n");
         holdingLeft = 0;
         mav_frameFn2Rmv(predictElbows, NULL);
         resetPosition();
         mav_frameFn2Add(trackElbow, NULL);
         }
       else // testing mode
         {
         if (!tracking) printf("warning:- not tracking, press 't' to switch tracking on\n");
         if (!netLoaded) printf("warning:- a network is not loaded, press '#' to load one\n");
         if (writing) // if writing, stop
           {
           mav_frameFn3Rmv(writeTrainingData, NULL); // stop any training
           writing = 0; // stop any writing
           closeTrainingFile();
           }
         printf("testing mode\n");
         holdingLeft = 0;
         resetPosition();
         mav_frameFn2Add(predictElbows, NULL);
         mav_frameFn2Add(trackElbow, NULL);
         }
       break;
     case ']':
       printf("scaling avatar, please enter file name: ");
       scanf("%s",&str1);
       scaleAvatar(str1);
       break;
     case '}':
       setOffsets();
       break;
     case '#':
       // load network
       printf("loading network, please enter file name: ");
       scanf("%s", &str1);
       netLoaded = loadNetwork(str1);
       break;
     case 't': // switch tracking
       switchTracking();
       break;
     case 'w': // change view
       switchView();
       break;
     case 'e': // switch between elbow position display
       elbowMode = (elbowMode+1)%4;
       avatar->holding_left = 0;
       avatar->holding_right = 0;
       if (elbowMode==1)
         {
         printf("keyboard only elbow position\n");
         // set arm position pointers
         leftArmRoll = &leftArmRollKeyboard;
         leftArmPitch = &leftArmPitchKeyboard;
         leftArmYaw = &leftArmYawKeyboard;
         rightArmRoll = &rightArmRollKeyboard;
         rightArmPitch = &rightArmPitchKeyboard;
         rightArmYaw = &rightArmYawKeyboard;
         if (programMode==NORMAL) holdingLeft = 1; // do not turn left arm on unless in normal use mode
         holdingRight = 1;
         }
       else if (elbowMode==2)
         {
         printf("network predicted elbow position\n");
         // set arm position pointers
         leftArmRoll = &leftArmRollNetwork;
         leftArmPitch = &leftArmPitchNetwork;
         leftArmYaw = &leftArmYawNetwork;
         rightArmRoll = &rightArmRollNetwork;
         rightArmPitch = &rightArmPitchNetwork;
         rightArmYaw = &rightArmYawNetwork;
         if (programMode==NORMAL) holdingLeft = 1;
         holdingRight = 1;
         }
       else if (elbowMode==3)
         {
         printf("tracked elbow position\n");
         // set arm position pointers
         leftArmRoll = &leftArmRollTracked;
         leftArmPitch = &leftArmPitchTracked;
         leftArmYaw = &leftArmYawTracked;
         rightArmRoll = &rightArmRollTracked;
         rightArmPitch = &rightArmPitchTracked;
         rightArmYaw = &rightArmYawTracked;
         if (programMode==NORMAL) holdingLeft = 1;
         holdingRight = 1;
         }
       else
         {
         printf("mechanical elbow position\n");
         holdingLeft = 0;
         holdingRight = 0;
         avatar->holding_left = 1;
         avatar->holding_right = 1;
         }
       break;
     case '|': // switch display of arm vectors
       vectorDisplay = (vectorDisplay+1)%6;
       if (vectorDisplay==1)
         {
         printf("displaying axis of avatar parts\n");
         mav_frameFn3Add(showAxis, NULL);
         }
       else if (vectorDisplay==2)
         {
         printf("displaying setArm vectors (predicted and corrected)\n");
         mav_frameFn3Rmv(showAxis, NULL);
         mav_frameFn3Add(showSetArmVectors, NULL);
         }
       else if (vectorDisplay==3)
         {
         printf("displaying upper body vectors\n");
         mav_frameFn3Rmv(showSetArmVectors, NULL);
         mav_frameFn3Add(showVectors, NULL);
         }
       else if (vectorDisplay==4)
         {
         printf("displaying upper body vectors and NN predicted elbow\n");
         mav_frameFn3Add(showNetworkElbow, NULL);
         }
       else if (vectorDisplay==5)
         {
         printf("displaying upper body vectors and tracked elbow\n");
         mav_frameFn3Rmv(showNetworkElbow, NULL);
         mav_frameFn3Add(showTrackedElbow, NULL);
         }
       else
         {
         printf("arm vector display off\n");
         mav_frameFn3Rmv(showTrackedElbow, NULL);
         mav_frameFn3Rmv(showVectors, NULL);
         }
       break;
     case '@': // set file names
       if (programMode==TRAINING) setTrainingFilenames();
       else if (programMode==TESTING) setTestingFilenames();
       else printf("cannot set filnames in normal mode:- press '-' to switch to training or testing mode\n");
       break;
     case '=': // start/stop writing data
       switchWriting();
       break;
     case '0':
       if (environment==1) deleteEnvironment1(&objects);
       else if (environment==2) deleteEnvironment2(&ground, &objects);
       else if (environment==3) deleteEnvironment3(&ground, &objects);
       environment = 0;
       break;
     case '1':
       if (environment==0) createEnvironment1(&objects);
       else if (environment==2)
         {
         deleteEnvironment2(&ground, &objects);
         createEnvironment1(&objects);
         }
       else if (environment==3)
         {
         deleteEnvironment3(&ground, &objects);
         createEnvironment1(&objects);
         }
       environment = 1;
       break;
     case '2':
       if (environment==0) createEnvironment2(&ground, &objects);
       else if (environment==1)
         {
         deleteEnvironment1(&objects);
         createEnvironment2(&ground, &objects);
         }
       else if (environment==3)
         {
         deleteEnvironment3(&ground, &objects);
         createEnvironment2(&ground, &objects);
         }
       environment = 2;
       break;
     case '3':
       if (environment==0) createEnvironment3(&ground, &objects);
       else if (environment==1)
         {
         deleteEnvironment1(&objects);
         createEnvironment3(&ground, &objects);
         }
       else if (environment==2)
         {
         deleteEnvironment2(&objects, &objects);
         createEnvironment3(&ground, &objects);
         }
       environment = 3;
       break;
     case '8':
       donkeyKong();
       break;
     case 'q': // exorcist quit
       eery = headYaw; // backup the head yaw (the exorcist function restores it)
       printf("aaeiiiiiiiiiiiiiiiii\n");
       mav_frameFn1Add(exorcist, NULL);
       break;
     }
   }

 // key switch for hold-key actions
 switch (keyEvent->key)
   {
   case 'h': // help
     if (keyEvent->movement == MAV_PRESSED)
       {
       // execute helpMessage function at end of frame (phase 3)
       help();
       mav_frameFn3Add(helpMessage, NULL);
       }
     else
       {
       // stop executing helpMessage function
       mav_frameFn3Rmv(helpMessage, NULL);
       }
     break;
   case ' ': // grab the object
     if (keyEvent->movement == MAV_PRESSED) // only consider key presses
       {
       pick(RIGHT_HAND);
       pick(LEFT_HAND);
       }
     else
       {
       drop(RIGHT_HAND);
       drop(LEFT_HAND);
       }
     break;
   case MAV_KEY_LEFT: // turn head to the left
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(headLeft, NULL);
     else mav_frameFn1Rmv(headLeft, NULL);
     break;
   case MAV_KEY_RIGHT: // turn head to the right
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(headRight, NULL);
     else mav_frameFn1Rmv(headRight, NULL);
     break;
   case MAV_KEY_UP: // tilt head up
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(headUp, NULL);
     else mav_frameFn1Rmv(headUp, NULL);
     break;
   case MAV_KEY_DOWN: // tilt head down
     if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(headDown, NULL);
     else mav_frameFn1Rmv(headDown, NULL);
     break;
   case '6': // hold left
     if (keyEvent->movement == MAV_PRESSED) pick(LEFT_HAND); //mav_frameFn1Add(pickLeft, NULL);
     else drop(LEFT_HAND); //mav_frameFn1Rmv(pickLeft, NULL);
     break;
   case '9': // hold right
     if (keyEvent->movement == MAV_PRESSED) pick(RIGHT_HAND); //mav_frameFn1Add(pickRight, NULL);
     else drop(RIGHT_HAND); //mav_frameFn1Rmv(pickRight, NULL);
     break;
   }

 // key switch for (mainly) hold-key actions that are dependant on keyboard control
 if (programMode==NORMAL)
   {
   switch (keyEvent->key)
     {
     case ('r'): // reset avatar position
       if (keyEvent->movement == MAV_PRESSED)
         {
         printf("resetting avatar position\n");
         resetPosition();
         }
       break;
     case 'y': // pitch torso forwards
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(torsoPitchForward, NULL);
       else mav_frameFn1Rmv(torsoPitchForward, NULL);
       break;
     case 'u': // pitch torso back
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(torsoPitchBack, NULL);
       else mav_frameFn1Rmv(torsoPitchBack, NULL);
       break;
     case 'i': // roll torso to the left
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(torsoRollLeft, NULL);
       else mav_frameFn1Rmv(torsoRollLeft, NULL);
       break;
     case 'o': // roll torso to the right
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(torsoRollRight, NULL);
       else mav_frameFn1Rmv(torsoRollRight, NULL);
       break;
     case 'p': // yaw torso right
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(torsoYawLeft, NULL);
       else mav_frameFn1Rmv(torsoYawLeft, NULL);
       break;
     case '[': // yaw torso left
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(torsoYawRight, NULL);
       else mav_frameFn1Rmv(torsoYawRight, NULL);
       break;
       
     case 'd': // left hand further away from body
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandFurther, NULL); // execute leftHandFurther function at phase 1
       else mav_frameFn1Rmv(leftHandFurther, NULL); // stop executing leftHandFurther function
       break;
     case 'f': // left hand closer in to body
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandCloser, NULL);
       else mav_frameFn1Rmv(leftHandCloser, NULL);
       break;
     case 'g': // left hand up
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandUp, NULL);
       else mav_frameFn1Rmv(leftHandUp, NULL);
       break;
     case 'b': // left hand down
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandDown, NULL);
       else mav_frameFn1Rmv(leftHandDown, NULL);
       break;
     case 'v': // left hand right
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandRight, NULL);
       else mav_frameFn1Rmv(leftHandRight, NULL);
       break;
     case 'c': // left hand left
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandLeft, NULL);
       else mav_frameFn1Rmv(leftHandLeft, NULL);
       break;
     case 'l': // right hand further away from body
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandFurther, NULL);
       else mav_frameFn1Rmv(rightHandFurther, NULL);
       break;
     case 'k': // right hand closer in to body
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandCloser, NULL);
       else mav_frameFn1Rmv(rightHandCloser, NULL);
       break;
     case 'j': // right hand up
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandUp, NULL);
       else mav_frameFn1Rmv(rightHandUp, NULL);
       break;
     case 'n': // right hand down
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandDown, NULL);
       else mav_frameFn1Rmv(rightHandDown, NULL);
       break;
     case ',': // right hand right
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandRight, NULL);
       else mav_frameFn1Rmv(rightHandRight, NULL);
       break;
     case 'm': // right hand left
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandLeft, NULL);
       else mav_frameFn1Rmv(rightHandLeft, NULL);
       break;

     case 'D': // left hand roll left
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandRollLeft, NULL); 
       else mav_frameFn1Rmv(leftHandRollLeft, NULL); 
       break;
     case 'F': // left hand roll right
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandRollRight, NULL);
       else mav_frameFn1Rmv(leftHandRollRight, NULL);
       break;
     case 'G': // left hand pich up
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandPitchUp, NULL);
       else mav_frameFn1Rmv(leftHandPitchUp, NULL);
       break;
     case 'B': // left hand pitch down
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandPitchDown, NULL);
       else mav_frameFn1Rmv(leftHandPitchDown, NULL);
       break;
     case 'V': // left hand yaw right
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandYawRight, NULL);
       else mav_frameFn1Rmv(leftHandYawRight, NULL);
       break;
     case 'C': // left hand yaw left
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftHandYawLeft, NULL);
       else mav_frameFn1Rmv(leftHandYawLeft, NULL);
       break;
     case 'L': // right hand roll right
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandRollRight, NULL);
       else mav_frameFn1Rmv(rightHandRollRight, NULL);
       break;
     case 'K': // right hand roll left
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandRollLeft, NULL);
       else mav_frameFn1Rmv(rightHandRollLeft, NULL);
       break;
     case 'J': // right hand pitch up
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandPitchUp, NULL);
       else mav_frameFn1Rmv(rightHandPitchUp, NULL);
       break;
     case 'N': // right hand pitch down
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandPitchDown, NULL);
       else mav_frameFn1Rmv(rightHandPitchDown, NULL);
       break;
     case '<': // right hand yaw right
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandYawRight, NULL);
       else mav_frameFn1Rmv(rightHandYawRight, NULL);
       break;
     case 'M': // right hand yaw left
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightHandYawLeft, NULL);
       else mav_frameFn1Rmv(rightHandYawLeft, NULL);
       break;
            
     case 's': // left arm pitch up
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftArmPitchUp, NULL);
       else mav_frameFn1Rmv(leftArmPitchUp, NULL);
       break;
     case 'x': // left arm pitch down
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftArmPitchDown, NULL);
       else mav_frameFn1Rmv(leftArmPitchDown, NULL);
       break;
     case 'a': // left arm yaw left
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftArmYawLeft, NULL);
       else mav_frameFn1Rmv(leftArmYawLeft, NULL);
       break;
     case 'z': // left arm yaw right
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(leftArmYawRight, NULL);
       else mav_frameFn1Rmv(leftArmYawRight, NULL);
       break;
     case ';': // right arm pitch up
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightArmPitchUp, NULL);
       else mav_frameFn1Rmv(rightArmPitchUp, NULL);
       break;
     case '.': // right arm pitch down
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightArmPitchDown, NULL);
       else mav_frameFn1Rmv(rightArmPitchDown, NULL);
       break;
     case '/': // left arm yaw left
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightArmYawLeft, NULL);
       else mav_frameFn1Rmv(rightArmYawLeft, NULL);
       break;
     case '\'': // left arm yaw right
       if (keyEvent->movement == MAV_PRESSED) mav_frameFn1Add(rightArmYawRight, NULL);
       else mav_frameFn1Rmv(rightArmYawRight, NULL);
       break;
     }
   }

 return 0;
 }

/**************************************************************************************************/

/***************************/
/* create main environment */
/***************************/
void createMainEnvironment(MAV_SMS **groundPlane, MAV_SMS **theObjects)
 {
 MAV_object *obj; // you need to define a MAV_object so that objects (such as a MAV_box) can be registered with the SMS

 printf("creating main environment...");

 // define a texture map from a file, texture index 5.
 mav_paletteTextureSet(mav_palette_default, 5, "marble_floor.ppm");

 // define objects
 defineRectangle(&rect);

 // register the objects as a maverik object (so you can register it with the SMS).
 // you must specify the class of the object.
 obj = mav_objectNew(mav_class_rectangle, &rect);
 // create the ground plane SMS and add a rectangle object to it
 *groundPlane = mav_SMSObjListNew();
 mav_SMSObjectAdd(*groundPlane, obj);
 // make objects in groundPlane SMS unselectable to keyboard and mouse events
 mav_SMSSelectabilitySet(*groundPlane, mav_win_current, MAV_FALSE);

 // create the main SMS and register the objects with it
 *theObjects = mav_SMSObjListNew();

 printf("done\n");
 }

/************************/
/* create environment 1 */
/************************/
void createEnvironment1(MAV_SMS **theObjects)
 {
 MAV_vector chestPos; // initial position of avatar's chest
 int count;
 float x,y,z;

 printf("creating environment 1...");
 chestPos = mav_vectorSet(0,1.4,0);
 // define and position spheres
 count = 0;
 for (x=-3; x<=3; x++)
   {
   for (y=-3; y<=3; y++)
     {
     for (z=-3; z<=3; z++)
       {
       setSphere(&sphere[count], 0.1, mav_vectorAdd(chestPos, mav_vectorSet(x/3,y/3,z/3)));
       mav_SMSObjectAdd(*theObjects, mav_objectNew(mav_class_sphere, &sphere[count]));
       count++;
       }
     }
   }
 // add function to delete spheres on contact with avatar's hands
 mav_frameFn2Add(deleteOnIntersect, NULL);
 printf("done\n");
 }

void deleteEnvironment1(MAV_SMS **theObjects)
 {
 printf("deleting environment 1...");
 mav_SMSCallbackEmptyExec(*theObjects, MAV_TRUE);
 mav_frameFn2Rmv(deleteOnIntersect, NULL);
 printf("done\n");
 }

/************************/
/* create environment 2 */
/************************/
void createEnvironment2(MAV_SMS **groundPlane, MAV_SMS **theObjects)
 {
 MAV_vector chestPos; // initial position of avatar's chest
 float x, y, z;
 int i;

 printf("creating environment 2...");
 chestPos = mav_vectorSet(0,1.4,0);
 // define and position cylinders
 for (i=0; i<5; i++)
   {
   setCylinder(&cyl[i], mav_vectorAdd(chestPos, mav_vectorSet((i*0.15)-0.3, 0, -0.50)));
   mav_SMSObjectAdd(*theObjects, mav_objectNew(mav_class_cylinder, &cyl[i]));
   }
 // create black hole
 setSphere(&hole, 0.025, mav_vectorAdd(chestPos, mav_vectorSet(0,0,-0.25)));
 hole.sp = mav_surfaceParamsNew(MAV_MATERIAL, 0, 5, 0);
 hole.nverts = 30;
 mav_SMSObjectAdd(*groundPlane, mav_objectNew(mav_class_sphere, &hole));
 // add function to delete object on contact with black hole
 mav_frameFn2Add(blackHole, NULL);
 printf("done\n");
 }

void deleteEnvironment2(MAV_SMS **groundPlane, MAV_SMS **theObjects)
 {
 printf("deleting environment 2...");
 mav_frameFn2Rmv(blackHole, NULL);
 mav_SMSCallbackEmptyExec(*theObjects, MAV_TRUE); // remove cylinders
 mav_SMSObjectRmv(*groundPlane, mav_objectNew(mav_class_sphere, &hole));
 printf("done\n");
 }

/************************/
/* create environment 3 */
/************************/
void createEnvironment3(MAV_SMS **groundPlane, MAV_SMS **theObjects)
 {
 MAV_vector chestPos; // initial position of avatar's chest
 float x, y, z;
 int i;

 printf("creating environment 3...");
 chestPos = mav_vectorSet(0,1.6,0);
 // define and position cylinders
 for (i=0; i<5; i++)
   {
   setCylinder(&cyl[i], mav_vectorAdd(chestPos, mav_vectorSet((i*0.15)-0.3, 0, -0.50)));
   mav_SMSObjectAdd(*theObjects, mav_objectNew(mav_class_cylinder, &cyl[i]));
   }
 // create black hole
 setSphere(&hole, 0.025, mav_vectorAdd(chestPos, mav_vectorSet(0,0,-0.25)));
 hole.sp = mav_surfaceParamsNew(MAV_MATERIAL, 0, 5, 0);
 hole.nverts = 30;
 mav_SMSObjectAdd(*groundPlane, mav_objectNew(mav_class_sphere, &hole));
 // add function to delete object on contact with black hole
 mav_frameFn2Add(blackHole, NULL);
 printf("done\n");
 }

void deleteEnvironment3(MAV_SMS **groundPlane, MAV_SMS **theObjects)
 {
 printf("deleting environment 3...");
 mav_frameFn2Rmv(blackHole, NULL);
 mav_SMSCallbackEmptyExec(*theObjects, MAV_TRUE); // remove cylinders
 mav_SMSObjectRmv(*groundPlane, mav_objectNew(mav_class_sphere, &hole));
 printf("done\n");
 }

/**************************************************************************************************/

/******************/
/* render a frame */
/******************/
// takes a Spatial Management Structure as it's arguement (with which objects register themselves)
void renderFrame(MAV_SMS *sms1, MAV_SMS *sms2)
 {
 // check for events
 mav_eventsCheck();

 // request start of new frame
 mav_frameBegin();

//  showAxis();

  // display SMS (in all windows)
  mav_SMSDisplay(mav_win_all, sms1);
  mav_SMSDisplay(mav_win_all, sms2);

 // request end of frame
 mav_frameEnd(); 
 }

/********/
/* main */
/********/
int main(int argc, char *argv[])
 {
// MAV_SMS *objects;
 MAV_viewParams vp; // view parameters
 int userLoaded = 0; // indicates whether the avatar has been scaled (for warning message)
 int trainFilenameLoaded = 0; // indicates whether a training data file name has been specified
 int testFilenameLoaded = 0; // ditto testing file name

 // check for help string
 if (argc==2)
   {
   if (!strcmp(argv[1],"-h") || !strcmp(argv[1],"-help"))
     {
     printUsage();
     exit(0);
     }
   }

 // turn on multi-sampling
 mav_opt_multiSample = MAV_TRUE;

 // set stereo vision
 mav_opt_stereo = MAV_STEREO_TWO_WINS;

 // set window sizes
 mav_opt_x = 0;
 mav_opt_y = 1152-480;
 mav_opt_width = 640;
 mav_opt_height = 480;
 mav_opt_right_x = 640;
 mav_opt_right_y = 1152-480;
 mav_opt_right_width = 640;
 mav_opt_right_height = 480;

 // specify TDM library
 mav_opt_TDMLib= "/opt/aig/TDM/v2.2b1/bary/lib/libtdm-polhemus.so";

 // turn on back-face culling (so you don't see inside of head if eye point is inside it)
// mav_windowBackfaceCullSet(mav_win_all, MAV_TRUE);

 // initialise maverik, avatar module, and TDM module (this dynamically loads the required libraries)
 mav_initialise(&argc, argv);
 mav_avatarModuleInit();
 mav_TDMModuleInit();

 // create and populate the environment
 createMainEnvironment(&ground, &objects);
 createAvatar(&ground);

 // check switches to set program mode
 if (argc==1)
   {
   programMode = NORMAL;
   }
 else if (argc>=2)
   {
   if (!strcmp(argv[1],"-n") || !strcmp(argv[1],"-normal"))
     {
     programMode = NORMAL;
     }
   else if (!strcmp(argv[1],"-t") || !strcmp(argv[1],"-train"))
     {
     programMode = TRAINING;
     }
   else if (!strcmp(argv[1],"-e") || !strcmp(argv[1],"-test"))
     {
     programMode = TESTING;
     }
   }// will now be in a program mode, so...
 if (argc>=3) //...check filenames
   {
   if (programMode==NORMAL)
     {
     if (scaleAvatar(argv[2])) userLoaded = 1;
     if (argc>3)
       {
       if (loadNetwork(argv[3])) netLoaded = 1;
       }
     }
   else if (programMode==TRAINING) // if training mode
     {
     if (scaleAvatar(argv[2])) userLoaded = 1;
     if (argc>3)
       {
       strcpy(trainingFilename, argv[3]);
       printf("training data file name set to '%s'\n", trainingFilename);
       trainFilenameLoaded = 1;
       }
     }
   else if (programMode==TESTING)
     {
     if (scaleAvatar(argv[2])) userLoaded = 1;
     if (argc>3)
       {
       if (loadNetwork(argv[3])) netLoaded = 1;
       }
     if (argc>4)
       {
       strcpy(testingFilename, argv[4]);
       printf("testing data file name set to '%s'\n", testingFilename);
       testFilenameLoaded = 1;
       }
     }
   }

 // print warning messages
 if (programMode==NORMAL)
   {
   printf("normal use mode\n");
   if (!userLoaded)
     printf("warning:- a user has not been specified, press ']' to scale avatar\n");
   if (!netLoaded)
     printf("warning:- a network has not been loaded, press '#' to load one\n");
   }
 if (programMode==TRAINING)
   {
   printf("training mode\n");
   if (!userLoaded)
     printf("warning:- a user has not been specified, press ']' to scale avatar\n");
   if (!trainFilenameLoaded)
     printf("warning:- a training data file has not been specified\n");
   if (!tracking)
     printf("warning:- not tracking, press 't' to start\n");
   }
 else if (programMode==TESTING)
   {
   printf("testing mode\n");
   if (!userLoaded)
     printf("warning:- a user has not been specified, press ']' to scale avatar\n");
   if (!netLoaded)
     printf("warning:- a network has not been loaded, press '#' to load one\n");
   if (!testFilenameLoaded)
     printf("warning:- a testing data file has not been specified\n");
   if (!tracking)
     printf("warning:- not tracking, press 't' to start\n");
   }

 // set rendering loop functions that are dependent on program mode
 if (programMode==NORMAL)
   mav_frameFn2Add(predictElbows, NULL);
 else if (programMode==TRAINING)
   mav_frameFn2Add(trackElbow, NULL);
 else if (programMode==TESTING)
   {
   mav_frameFn2Add(predictElbows, NULL);
   mav_frameFn2Add(trackElbow, NULL);
   }

 mav_stp_default.offset = 0.065; // set the stereo offset (must be done after initialising maverik)

 // set up viewing parameters (needed to prevent run-time warning)
 vp.eye = mav_matrixXYZGet(avatar->matrix);
 vp.view = mav_vectorSet(0,0,-1);
 vp.right = MAV_X_VECTOR;
 vp.up = MAV_Y_VECTOR;
 vp.mod = avatarView; // modify eye view using avatar head position

 // use these view parameters in all windows
 mav_windowViewParamsSet(mav_win_all, &vp);

 // default mouse navigation
 mav_navigationMouse(mav_win_all, mav_navigationMouseDefault);
 // customize behaviour
 mav_navigationMouseDefaultParams(mav_win_all, MAV_LEFT_BUTTON, turn, 0.01, -0.0005, walk, 0.01, 0.0005);
 mav_navigationMouseDefaultParams(mav_win_all, MAV_RIGHT_BUTTON, yawHead, 0.01, -0.0005, pitchHead, 0.01, -0.0005);

 // define keyboard event callback (the same for all windows, when the mouse is anywhere in the world)
 mav_callbackKeyboardSet(mav_win_all, mav_class_world, keyEventCallback);

 // define TDM callback
 mav_callbackTDMSet(mav_win_all, mav_class_world, tdmEventCallback);

 // start rendering loop
 while(1) renderFrame(ground,objects);

 return 0;
 }

/**************************************************************************************************/

